@@ -239,7 +239,9 @@ wxString get_command_path(const wxChar * command_name)
 }
 
 CavernLogWindow::CavernLogWindow(MainFrm * mainfrm_, const wxString & survey_, wxWindow * parent)
-    : wxHtmlWindow(parent), mainfrm(mainfrm_), cavern_out(NULL),
+    : wxHtmlWindow(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize,
+		   wxHW_DEFAULT_STYLE|wxHW_NO_SELECTION),
+      mainfrm(mainfrm_), cavern_out(NULL), highlight(NULL),
       link_count(0), end(buf), init_done(false), survey(survey_)
 #ifdef CAVERNLOG_USE_THREADS
       , thread(NULL)
@@ -372,7 +374,7 @@ CavernLogWindow::OnLinkClicked(const wxHtmlLinkInfo &link)
 	    }
 	    case wxT('c'): {
 		wxString l;
-		if (colon2 >= href.size())
+		if (colon2 >= href.size() - 1)
 		    l = wxT("0");
 		else
 		    l = escape_for_shell(href.substr(colon2 + 1));
@@ -530,6 +532,49 @@ bad_utf8:
 		    break;
 		case '\n': {
 		    if (cur.empty()) continue;
+		    if (cur[0] == ' ') {
+			if (source_line.empty()) {
+			    // Source line shown for context.  Store it so we
+			    // can use the caret line to highlight it.
+			    swap(source_line, cur);
+			} else {
+			    size_t caret = cur.rfind('^');
+			    if (caret != wxString::npos) {
+				size_t tilde = cur.rfind('~');
+				if (tilde == wxString::npos || tilde < caret) {
+				    tilde = caret;
+				}
+				cur = "&nbsp;";
+				// FIXME: Need to count each & entity as one character...
+				cur.append(source_line, 1, caret - 1);
+				cur.append("<b>");
+				cur.append(highlight ? highlight : wxT("<span \"color:green\">"));
+				cur.append(source_line, caret, tilde + 1 - caret);
+				cur.append("</span></b>");
+				cur.append(source_line, tilde + 1, wxString::npos);
+			    } else {
+				// No caret in second line - just output both.
+				source_line.replace(0, 1, "&nbsp;");
+				source_line += "<br>\n&nbsp;";
+				source_line.append(cur, 1, wxString::npos);
+				swap(cur, source_line);
+			    }
+			    cur += "<br>\n";
+			    AppendToPage(cur);
+			    cur.clear();
+			    source_line.clear();
+			}
+			continue;
+		    }
+
+		    if (!source_line.empty()) {
+			// Previous line was a source line without column info
+			// so just show it.
+			source_line.replace(0, 1, "&nbsp;");
+			source_line += "<br>\n";
+			AppendToPage(source_line);
+			source_line.clear();
+		    }
 #ifndef __WXMSW__
 		    size_t colon = cur.find(':');
 #else
@@ -549,7 +594,8 @@ bad_utf8:
 			    // Check for column number.
 			    while (++i < cur.size() - 2 &&
 			       cur[i] >= wxT('0') && cur[i] <= wxT('9')) { }
-			    if (i > colon + 1 && cur[i] == wxT(':') ) {
+			    bool have_column = (i > colon + 1 && cur[i] == wxT(':'));
+			    if (have_column) {
 				colon = i;
 			    } else {
 				// If there's no colon, include a trailing ':'
@@ -570,19 +616,25 @@ bad_utf8:
 			    cur.insert(offset, wxT("</a>"));
 			    offset += 4 + 2;
 
+			    if (!have_column) --offset;
+
 			    static const wxString & error_marker = wmsg(/*error*/93) + ":";
 			    static const wxString & warning_marker = wmsg(/*warning*/4) + ":";
 
 			    if (cur.substr(offset, error_marker.size()) == error_marker) {
 				// Show "error" marker in red.
-				cur.insert(offset, wxT("<span style=\"color:red\">"));
+				highlight = wxT("<span style=\"color:red\">");
+				cur.insert(offset, highlight);
 				offset += 24 + error_marker.size() - 1;
 				cur.insert(offset, wxT("</span>"));
 			    } else if (cur.substr(offset, warning_marker.size()) == warning_marker) {
 				// Show "warning" marker in orange.
-				cur.insert(offset, wxT("<span style=\"color:orange\">"));
+				highlight = wxT("<span style=\"color:orange\">");
+				cur.insert(offset, highlight);
 				offset += 27 + warning_marker.size() - 1;
 				cur.insert(offset, wxT("</span>"));
+			    } else {
+				highlight = NULL;
 			    }
 
 			    ++link_count;
@@ -628,11 +680,20 @@ bad_utf8:
 		    cur += wxT("&#22;");
 		    continue;
 		default:
+#ifdef wxUSE_UNICODE
+		    cur += wxChar(ch);
+#else
+		    // This approach means that highlighting of "error" or
+		    // "warning" won't work in translations where they contain
+		    // non-ASCII characters, but wxWidgets >= 3.0 in always
+		    // Unicode, so this corner case is already very uncommon,
+		    // and will become irrelevant with time.
 		    if (ch >= 128) {
 			cur += wxString::Format(wxT("&#%u;"), ch);
 		    } else {
 			cur += (char)ch;
 		    }
+#endif
 	    }
 	}
 
@@ -643,10 +704,23 @@ bad_utf8:
 	return;
     }
 
+    if (!source_line.empty()) {
+	// Previous line was a source line without column info
+	// so just show it.
+	source_line.replace(0, 1, "&nbsp;");
+	source_line += "<br>\n";
+	AppendToPage(source_line);
+	source_line.clear();
+    }
+
     if (e.len <= 0 && buf != end) {
 	// Truncated UTF-8 sequence.
 	cur += badutf8_html;
     }
+    if (!cur.empty()) {
+	cur += "<br>\n";
+	AppendToPage("<hr>" + cur);
+    }
 
     /* TRANSLATORS: Label for button in aven’s cavern log window which
      * allows the user to save the log to a file. */
