@@ -375,8 +375,7 @@ get_units(unsigned long qmask, bool percent_ok)
    get_token();
    units = match_tok(utab, TABSIZE(utab));
    if (units == UNITS_NULL) {
-      file.lpos += strlen(buffer);
-      compile_error_skip(-/*Unknown units “%s”*/35, buffer);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Unknown units “%s”*/35, buffer);
       return UNITS_NULL;
    }
    if (units == UNITS_PERCENT && percent_ok &&
@@ -385,11 +384,10 @@ get_units(unsigned long qmask, bool percent_ok)
    }
    if (((qmask & LEN_QMASK) && !TSTBIT(LEN_UMASK, units)) ||
        ((qmask & ANG_QMASK) && !TSTBIT(ANG_UMASK, units))) {
-      file.lpos += strlen(buffer);
       /* TRANSLATORS: Note: In English you talk about the *units* of a single
        * measurement, but the correct term in other languages may be singular.
        */
-      compile_error_skip(-/*Invalid units “%s” for quantity*/37, buffer);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Invalid units “%s” for quantity*/37, buffer);
       return UNITS_NULL;
    }
    return units;
@@ -453,17 +451,15 @@ get_qlist(unsigned long mask_bad)
       if (tok == Q_NULL) break;
       qmask |= BIT(tok);
       if (qmask & mask_bad) {
-	 file.lpos += strlen(buffer);
-	 compile_error_skip(-/*Unknown instrument “%s”*/39, buffer);
+	 compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Unknown instrument “%s”*/39, buffer);
 	 return 0;
       }
    }
 
    if (qmask == 0) {
-      file.lpos += strlen(buffer);
       /* TRANSLATORS: A "quantity" is something measured like "LENGTH",
        * "BEARING", "ALTITUDE", etc. */
-      compile_error_skip(-/*Unknown quantity “%s”*/34, buffer);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Unknown quantity “%s”*/34, buffer);
    } else {
       set_pos(&fp);
    }
@@ -500,27 +496,25 @@ cmd_set(void)
    mask = match_tok(chartab, TABSIZE(chartab));
 
    if (mask == SPECIAL_UNKNOWN) {
-      file.lpos += strlen(buffer);
-      compile_error_skip(-/*Unknown character class “%s”*/42, buffer);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Unknown character class “%s”*/42, buffer);
       return;
    }
 
 #ifndef NO_DEPRECATED
    if (mask == SPECIAL_ROOT) {
       if (root_depr_count < 5) {
-	 file.lpos += strlen(buffer);
 	 /* TRANSLATORS: Use of the ROOT character (which is "\" by default) is
 	  * deprecated, so this error would be generated by:
 	  *
 	  * *equate \foo.7 1
 	  *
 	  * If you're unsure what "deprecated" means, see:
-	  * http://en.wikipedia.org/wiki/Deprecation */
-	 compile_warning(-/*ROOT is deprecated*/25);
+	  * https://en.wikipedia.org/wiki/Deprecation */
+	 compile_diagnostic(DIAG_WARN|DIAG_BUF, /*ROOT is deprecated*/25);
 	 if (++root_depr_count == 5)
 	     /* TRANSLATORS: If you're unsure what "deprecated" means, see:
-	      * http://en.wikipedia.org/wiki/Deprecation */
-	    compile_warning(/*Further uses of this deprecated feature will not be reported*/95);
+	      * https://en.wikipedia.org/wiki/Deprecation */
+	    compile_diagnostic(DIAG_WARN, /*Further uses of this deprecated feature will not be reported*/95);
       }
    }
 #endif
@@ -569,16 +563,19 @@ cmd_set(void)
 }
 
 static void
-check_reentry(prefix *survey)
+check_reentry(prefix *survey, const filepos* fpos_ptr)
 {
    /* Don't try to check "*prefix \" or "*begin \" */
    if (!survey->up) return;
    if (TSTBIT(survey->sflags, SFLAGS_PREFIX_ENTERED)) {
       static int reenter_depr_count = 0;
+      filepos fp_tmp;
 
       if (reenter_depr_count >= 5)
 	 return;
 
+      get_pos(&fp_tmp);
+      set_pos(fpos_ptr);
       /* TRANSLATORS: The first of two warnings given when a survey which has
        * already been completed is reentered.  This example file crawl.svx:
        *
@@ -595,8 +592,9 @@ check_reentry(prefix *survey)
        * crawl.svx:1: Originally entered here
        *
        * If you're unsure what "deprecated" means, see:
-       * http://en.wikipedia.org/wiki/Deprecation */
-      compile_warning(/*Reentering an existing survey is deprecated*/29);
+       * https://en.wikipedia.org/wiki/Deprecation */
+      compile_diagnostic(DIAG_WARN|DIAG_TOKEN, /*Reentering an existing survey is deprecated*/29);
+      set_pos(&fp_tmp);
       /* TRANSLATORS: The second of two warnings given when a survey which has
        * already been completed is reentered.  This example file crawl.svx:
        *
@@ -613,15 +611,15 @@ check_reentry(prefix *survey)
        * crawl.svx:1: Originally entered here
        *
        * If you're unsure what "deprecated" means, see:
-       * http://en.wikipedia.org/wiki/Deprecation */
-      compile_warning_pfx(survey, /*Originally entered here*/30);
+       * https://en.wikipedia.org/wiki/Deprecation */
+      compile_diagnostic_pfx(DIAG_WARN, survey, /*Originally entered here*/30);
       if (++reenter_depr_count == 5) {
 	 /* After we've warned about 5 uses of the same deprecated feature, we
 	  * give up for the rest of the current processing run.
 	  *
 	  * If you're unsure what "deprecated" means, see:
-	  * http://en.wikipedia.org/wiki/Deprecation */
-	 compile_warning(/*Further uses of this deprecated feature will not be reported*/95);
+	  * https://en.wikipedia.org/wiki/Deprecation */
+	 compile_diagnostic(DIAG_WARN, /*Further uses of this deprecated feature will not be reported*/95);
       }
    } else {
       survey->sflags |= BIT(SFLAGS_PREFIX_ENTERED);
@@ -636,19 +634,21 @@ cmd_prefix(void)
 {
    static int prefix_depr_count = 0;
    prefix *survey;
+   filepos fp;
    /* Issue warning first, so "*prefix \" warns first that *prefix is
     * deprecated and then that ROOT is...
     */
    if (prefix_depr_count < 5) {
       /* TRANSLATORS: If you're unsure what "deprecated" means, see:
-       * http://en.wikipedia.org/wiki/Deprecation */
-      compile_warning(-/**prefix is deprecated - use *begin and *end instead*/6);
+       * https://en.wikipedia.org/wiki/Deprecation */
+      compile_diagnostic(DIAG_WARN|DIAG_BUF, /**prefix is deprecated - use *begin and *end instead*/6);
       if (++prefix_depr_count == 5)
-	 compile_warning(/*Further uses of this deprecated feature will not be reported*/95);
+	 compile_diagnostic(DIAG_WARN, /*Further uses of this deprecated feature will not be reported*/95);
    }
+   get_pos(&fp);
    survey = read_prefix(PFX_SURVEY|PFX_ALLOW_ROOT);
    pcs->Prefix = survey;
-   check_reentry(survey);
+   check_reentry(survey, &fp);
 }
 #endif
 
@@ -671,7 +671,7 @@ cmd_alias(void)
    pcs->dash_for_anon_wall_station = (*buffer != '\0');
    return;
 bad:
-   compile_error_skip(/*Bad *alias command*/397);
+   compile_diagnostic(DIAG_ERR|DIAG_SKIP, /*Bad *alias command*/397);
 }
 
 static void
@@ -688,10 +688,13 @@ cmd_begin(void)
    skipblanks();
    pcs->begin_survey = NULL;
    if (!isEol(ch) && !isComm(ch)) {
-      prefix *survey = read_prefix(PFX_SURVEY|PFX_ALLOW_ROOT|PFX_WARN_SEPARATOR);
+      filepos fp;
+      prefix *survey;
+      get_pos(&fp);
+      survey = read_prefix(PFX_SURVEY|PFX_ALLOW_ROOT|PFX_WARN_SEPARATOR);
       pcs->begin_survey = survey;
       pcs->Prefix = survey;
-      check_reentry(survey);
+      check_reentry(survey, &fp);
       f_export_ok = fTrue;
    }
 }
@@ -723,15 +726,16 @@ cmd_end(void)
 {
    settings *pcsParent;
    prefix *survey, *begin_survey;
+   filepos fp;
 
    pcsParent = pcs->next;
 
    if (pcs->begin_lineno == 0) {
       if (pcsParent == NULL) {
 	 /* more ENDs than BEGINs */
-	 compile_error_skip(/*No matching BEGIN*/192);
+	 compile_diagnostic(DIAG_ERR|DIAG_SKIP, /*No matching BEGIN*/192);
       } else {
-	 compile_error_skip(/*END with no matching BEGIN in this file*/22);
+	 compile_diagnostic(DIAG_ERR|DIAG_SKIP, /*END with no matching BEGIN in this file*/22);
       }
       return;
    }
@@ -747,11 +751,13 @@ cmd_end(void)
    if (isEol(ch) || isComm(ch)) {
       survey = NULL;
    } else {
+      get_pos(&fp);
       survey = read_prefix(PFX_SURVEY|PFX_ALLOW_ROOT);
    }
 
    if (survey != begin_survey) {
       if (survey) {
+	 set_pos(&fp);
 	 if (!begin_survey) {
 	    /* TRANSLATORS: Used when a BEGIN command has no survey, but the
 	     * END command does, e.g.:
@@ -759,12 +765,13 @@ cmd_end(void)
 	     * *begin
 	     * 1 2 10.00 178 -01
 	     * *end entrance      <--[Message given here] */
-	    compile_error_skip(-/*Matching BEGIN command has no survey name*/36);
+	    compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Matching BEGIN command has no survey name*/36);
 	 } else {
 	    /* TRANSLATORS: *BEGIN <survey> and *END <survey> should have the
 	     * same <survey> if it’s given at all */
-	    compile_error_skip(-/*Survey name doesn’t match BEGIN*/193);
+	    compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Survey name doesn’t match BEGIN*/193);
 	 }
+	 skipline();
       } else {
 	 /* TRANSLATORS: Used when a BEGIN command has a survey name, but the
 	  * END command omits it, e.g.:
@@ -772,7 +779,7 @@ cmd_end(void)
 	  * *begin entrance
 	  * 1 2 10.00 178 -01
 	  * *end     <--[Message given here] */
-	 compile_warning(-/*Survey name omitted from END*/194);
+	 compile_diagnostic(DIAG_WARN|DIAG_COL, /*Survey name omitted from END*/194);
       }
    }
 }
@@ -819,26 +826,27 @@ cmd_fix(void)
    }
    if (x == HUGE_REAL) {
       if (pcs->proj || proj_out) {
-	 compile_error_skip(/*Coordinates can't be omitted when coordinate system has been specified*/439);
+	 compile_diagnostic(DIAG_ERR|DIAG_COL|DIAG_SKIP, /*Coordinates can't be omitted when coordinate system has been specified*/439);
 	 return;
       }
 
       if (fix_name == name_omit_already) {
-	 compile_warning(/*Same station fixed twice with no coordinates*/61);
+	 compile_diagnostic(DIAG_WARN|DIAG_COL, /*Same station fixed twice with no coordinates*/61);
 	 return;
       }
 
       /* TRANSLATORS: " *fix a " gives this message: */
-      compile_warning(/*FIX command with no coordinates - fixing at (0,0,0)*/54);
+      compile_diagnostic(DIAG_WARN|DIAG_COL, /*FIX command with no coordinates - fixing at (0,0,0)*/54);
 
       if (name_omit_already) {
 	 /* TRANSLATORS: Emitted after second and subsequent "FIX command with
 	  * no coordinates - fixing at (0,0,0)" warnings.
 	  */
-	 compile_error_at(name_omit_already_filename,
-			  name_omit_already_line,
-			  /*Already had FIX command with no coordinates for station “%s”*/441,
-			  sprint_prefix(name_omit_already));
+	 compile_diagnostic_at(DIAG_ERR|DIAG_COL,
+			       name_omit_already_filename,
+			       name_omit_already_line,
+			       /*Already had FIX command with no coordinates for station “%s”*/441,
+			       sprint_prefix(name_omit_already));
       } else {
 	 name_omit_already = fix_name;
 	 name_omit_already_filename = file.filename;
@@ -859,31 +867,36 @@ cmd_fix(void)
 	 }
 	 int r = pj_transform(pcs->proj, proj_out, 1, 1, &x, &y, &z);
 	 if (r != 0) {
-	    compile_error(/*Failed to convert coordinates: %s*/436, pj_strerrno(r));
+	    compile_diagnostic(DIAG_ERR, /*Failed to convert coordinates: %s*/436, pj_strerrno(r));
 	 }
       } else if (pcs->proj) {
-	 compile_error(/*The input projection is set but the output projection isn't*/437);
+	 compile_diagnostic(DIAG_ERR, /*The input projection is set but the output projection isn't*/437);
       } else if (proj_out) {
-	 compile_error(/*The output projection is set but the input projection isn't*/438);
+	 compile_diagnostic(DIAG_ERR, /*The output projection is set but the input projection isn't*/438);
       }
 
+      get_pos(&fp);
       sdx = read_numeric(fTrue);
       if (sdx <= 0) {
-	  compile_error_skip(-/*Standard deviation must be positive*/48);
+	  set_pos(&fp);
+	  compile_diagnostic(DIAG_ERR|DIAG_SKIP|DIAG_NUM, /*Standard deviation must be positive*/48);
 	  return;
       }
       if (sdx != HUGE_REAL) {
 	 real sdy, sdz;
 	 real cxy = 0, cyz = 0, czx = 0;
+	 get_pos(&fp);
 	 sdy = read_numeric(fTrue);
 	 if (sdy == HUGE_REAL) {
 	    /* only one variance given */
 	    sdy = sdz = sdx;
 	 } else {
 	    if (sdy <= 0) {
-	       compile_error_skip(-/*Standard deviation must be positive*/48);
+	       set_pos(&fp);
+	       compile_diagnostic(DIAG_ERR|DIAG_SKIP|DIAG_NUM, /*Standard deviation must be positive*/48);
 	       return;
 	    }
+	    get_pos(&fp);
 	    sdz = read_numeric(fTrue);
 	    if (sdz == HUGE_REAL) {
 	       /* two variances given - horizontal & vertical */
@@ -891,7 +904,8 @@ cmd_fix(void)
 	       sdy = sdx;
 	    } else {
 	       if (sdz <= 0) {
-		  compile_error_skip(-/*Standard deviation must be positive*/48);
+		  set_pos(&fp);
+		  compile_diagnostic(DIAG_ERR|DIAG_SKIP|DIAG_NUM, /*Standard deviation must be positive*/48);
 		  return;
 	       }
 	       cxy = read_numeric(fTrue);
@@ -968,11 +982,11 @@ cmd_fix(void)
    }
 
    if (x != POS(stn, 0) || y != POS(stn, 1) || z != POS(stn, 2)) {
-      compile_error(/*Station already fixed or equated to a fixed point*/46);
+      compile_diagnostic(DIAG_ERR, /*Station already fixed or equated to a fixed point*/46);
       return;
    }
    /* TRANSLATORS: *fix a 1 2 3 / *fix a 1 2 3 */
-   compile_warning(/*Station already fixed at the same coordinates*/55);
+   compile_diagnostic(DIAG_WARN, /*Station already fixed at the same coordinates*/55);
 }
 
 static void
@@ -998,8 +1012,7 @@ cmd_flags(void)
       flag = match_tok(flagtab, TABSIZE(flagtab));
       /* treat the second NOT in "NOT NOT" as an unknown flag */
       if (flag == FLAGS_UNKNOWN || (fNot && flag == FLAGS_NOT)) {
-	 file.lpos += strlen(buffer);
-	 compile_error(-/*FLAG “%s” unknown*/68, buffer);
+	 compile_diagnostic(DIAG_ERR|DIAG_BUF, /*FLAG “%s” unknown*/68, buffer);
 	 /* Recover from “*FLAGS NOT BOGUS SURFACE” by ignoring "NOT BOGUS" */
 	 fNot = fFalse;
       } else if (flag == FLAGS_NOT) {
@@ -1013,11 +1026,9 @@ cmd_flags(void)
    }
 
    if (fNot) {
-      file.lpos += strlen(buffer);
-      compile_error(-/*Expecting “DUPLICATE”, “SPLAY”, or “SURFACE”*/188);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF, /*Expecting “DUPLICATE”, “SPLAY”, or “SURFACE”*/188);
    } else if (fEmpty) {
-      file.lpos += strlen(buffer);
-      compile_error(-/*Expecting “NOT”, “DUPLICATE”, “SPLAY”, or “SURFACE”*/189);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF, /*Expecting “NOT”, “DUPLICATE”, “SPLAY”, or “SURFACE”*/189);
    }
 }
 
@@ -1026,19 +1037,22 @@ cmd_equate(void)
 {
    prefix *name1, *name2;
    bool fOnlyOneStn = fTrue; /* to trap eg *equate entrance.6 */
+   filepos fp;
 
+   get_pos(&fp);
    name1 = read_prefix(PFX_STATION|PFX_ALLOW_ROOT|PFX_SUSPECT_TYPO);
    while (fTrue) {
       name2 = name1;
       skipblanks();
       if (isEol(ch) || isComm(ch)) {
 	 if (fOnlyOneStn) {
+	    set_pos(&fp);
 	    /* TRANSLATORS: EQUATE is a command name, so shouldn’t be
 	     * translated.
 	     *
 	     * Here "station" is a survey station, not a train station.
 	     */
-	    compile_error_skip(-/*Only one station in EQUATE command*/33);
+	    compile_diagnostic(DIAG_ERR|DIAG_SKIP|DIAG_TOKEN, /*Only one station in EQUATE command*/33);
 	 }
 	 return;
       }
@@ -1078,10 +1092,10 @@ report_missing_export(prefix *pfx, int depth)
        * Here "survey" is a "cave map" rather than list of questions - it should be
        * translated to the terminology that cavers using the language would use.
        */
-      compile_error_pfx(survey,
-			/*Station “%s” not exported from survey “%s”*/26, p, s);
+      compile_diagnostic_pfx(DIAG_ERR, survey,
+			     /*Station “%s” not exported from survey “%s”*/26, p, s);
    } else {
-      compile_error(/*Station “%s” not exported from survey “%s”*/26, p, s);
+      compile_diagnostic(DIAG_ERR, /*Station “%s” not exported from survey “%s”*/26, p, s);
    }
    osfree(s);
 }
@@ -1133,8 +1147,8 @@ cmd_export(void)
 	     * *export 1
 	     * 1 2 1.24 045 -6
 	     * *end example */
-	    compile_error(/*Station “%s” already exported*/66,
-			  sprint_prefix(pfx));
+	    compile_diagnostic(DIAG_ERR, /*Station “%s” already exported*/66,
+			       sprint_prefix(pfx));
 	 }
 	 pfx->min_export = depth;
       }
@@ -1263,9 +1277,8 @@ cmd_data(void)
    }
 
    if (style == STYLE_UNKNOWN) {
-      file.lpos += strlen(buffer);
       /* TRANSLATORS: e.g. trying to refer to an invalid FNORD data style */
-      compile_error_skip(-/*Data style “%s” unknown*/65, buffer);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Data style “%s” unknown*/65, buffer);
       return;
    }
 
@@ -1276,11 +1289,10 @@ cmd_data(void)
    if (isOmit(ch)) {
       static int data_depr_count = 0;
       if (data_depr_count < 5) {
-	 file.lpos += strlen(buffer);
-	 compile_warning(-/*“*data %s %c …” is deprecated - use “*data %s …” instead*/104,
-			 buffer, ch, buffer);
+	 compile_diagnostic(DIAG_WARN|DIAG_BUF, /*“*data %s %c …” is deprecated - use “*data %s …” instead*/104,
+			    buffer, ch, buffer);
 	 if (++data_depr_count == 5)
-	    compile_warning(/*Further uses of this deprecated feature will not be reported*/95);
+	    compile_diagnostic(DIAG_WARN, /*Further uses of this deprecated feature will not be reported*/95);
       }
       nextch();
    }
@@ -1299,26 +1311,26 @@ cmd_data(void)
       }
       /* Note: an unknown token is reported as trailing garbage */
       if (!TSTBIT(mask_all[style], d)) {
-	 file.lpos += strlen(buffer);
 	 /* TRANSLATORS: a data "style" is something like NORMAL, DIVING, etc.
 	  * a "reading" is one of FROM, TO, TAPE, COMPASS, CLINO for NORMAL
 	  * neither style nor reading is a keyword in the program This error
 	  * complains about a depth gauge reading in normal style, for example
 	  */
-	 compile_error_skip(-/*Reading “%s” not allowed in data style “%s”*/63,
-		       buffer, style_name);
+	 compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP,
+			    /*Reading “%s” not allowed in data style “%s”*/63,
+			    buffer, style_name);
 	 osfree(style_name);
 	 osfree(new_order);
 	 return;
       }
       if (TSTBIT(mUsed, Newline) && TSTBIT(m_multi, d)) {
-	 file.lpos += strlen(buffer);
 	 /* TRANSLATORS: caused by e.g.
 	  *
 	  * *data diving station newline depth tape compass
 	  *
 	  * ("depth" needs to occur before "newline"). */
-	 compile_error_skip(-/*Reading “%s” must occur before NEWLINE*/225, buffer);
+	 compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP,
+			    /*Reading “%s” must occur before NEWLINE*/225, buffer);
 	 osfree(style_name);
 	 osfree(new_order);
 	 return;
@@ -1328,10 +1340,9 @@ cmd_data(void)
        */
       if (!((BIT(Ignore) | BIT(End) | BIT(IgnoreAll)) & BIT(d))) {
 	 if (TSTBIT(mUsed, d)) {
-	    file.lpos += strlen(buffer);
 	    /* TRANSLATORS: complains about a situation like trying to define
 	     * two from stations per leg */
-	    compile_error_skip(-/*Duplicate reading “%s”*/67, buffer);
+	    compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Duplicate reading “%s”*/67, buffer);
 	    osfree(style_name);
 	    osfree(new_order);
 	    return;
@@ -1367,21 +1378,19 @@ cmd_data(void)
 	       break;
 	     case Newline:
 	       if (mUsed & ~m_multi) {
-		  file.lpos += strlen(buffer);
 		  /* TRANSLATORS: e.g.
 		   *
 		   * *data normal from to tape newline compass clino */
-		  compile_error_skip(-/*NEWLINE can only be preceded by STATION, DEPTH, and COUNT*/226);
+		  compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*NEWLINE can only be preceded by STATION, DEPTH, and COUNT*/226);
 		  osfree(style_name);
 		  osfree(new_order);
 		  return;
 	       }
 	       if (k == 0) {
-		  file.lpos += strlen(buffer);
 		  /* TRANSLATORS: error from:
 		   *
 		   * *data normal newline from to tape compass clino */
-		  compile_error_skip(-/*NEWLINE can’t be the first reading*/222);
+		  compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*NEWLINE can’t be the first reading*/222);
 		  osfree(style_name);
 		  osfree(new_order);
 		  return;
@@ -1392,14 +1401,13 @@ cmd_data(void)
 	    }
 	    if (fBad) {
 	       /* Not entirely happy with phrasing this... */
-	       file.lpos += strlen(buffer);
 	       /* TRANSLATORS: This is an error from the *DATA command.  It
 		* means that a reading (which will appear where %s is isn't
 		* valid as the list of readings has already included the same
 		* reading, or an equivalent one (e.g. you can't have both
 		* DEPTH and DEPTHCHANGE together). */
-	       compile_error_skip(-/*Reading “%s” duplicates previous reading(s)*/77,
-			     buffer);
+	       compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Reading “%s” duplicates previous reading(s)*/77,
+					 buffer);
 	       osfree(style_name);
 	       osfree(new_order);
 	       return;
@@ -1420,11 +1428,10 @@ cmd_data(void)
    } while (d != End);
 
    if (k >= 2 && new_order[k - 2] == Newline) {
-      file.lpos += strlen(buffer);
       /* TRANSLATORS: error from:
        *
        * *data normal from to tape compass clino newline */
-      compile_error_skip(-/*NEWLINE can’t be the last reading*/223);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*NEWLINE can’t be the last reading*/223);
       osfree(style_name);
       osfree(new_order);
       return;
@@ -1447,7 +1454,7 @@ cmd_data(void)
        * *data normal station tape compass clino
        *
        * ("station" signifies interleaved data). */
-      compile_error_skip(/*Interleaved readings, but no NEWLINE*/224);
+      compile_diagnostic(DIAG_ERR|DIAG_SKIP, /*Interleaved readings, but no NEWLINE*/224);
       osfree(style_name);
       osfree(new_order);
       return;
@@ -1491,7 +1498,7 @@ cmd_data(void)
       SVX_ASSERT((((mUsed &~ BIT(Newline)) | mask_optional[style])
 	      &~ mask[style]) == 0);
       /* TRANSLATORS: i.e. not enough readings for the style. */
-      compile_error_skip(/*Too few readings for data style “%s”*/64, style_name);
+      compile_diagnostic(DIAG_ERR|DIAG_SKIP, /*Too few readings for data style “%s”*/64, style_name);
       osfree(style_name);
       osfree(new_order);
       return;
@@ -1523,6 +1530,7 @@ cmd_units(void)
    unsigned long qmask;
    unsigned long m; /* mask with bit x set to indicate quantity x specified */
    real factor;
+   filepos fp;
 
    qmask = get_qlist(BIT(Q_POS)|BIT(Q_PLUMB)|BIT(Q_LEVEL));
 
@@ -1532,13 +1540,16 @@ cmd_units(void)
       return;
    }
 
+   get_pos(&fp);
    factor = read_numeric(fTrue);
    if (factor == 0.0) {
+      set_pos(&fp);
       /* TRANSLATORS: error message given by "*units tape 0 feet" - it’s
        * meaningless to say your tape is marked in "0 feet" (but you might
        * measure distance by counting knots on a diving line, and tie them
        * every "2 feet"). */
-      compile_error_skip(-/**UNITS factor must be non-zero*/200);
+      compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /**UNITS factor must be non-zero*/200);
+      skipline();
       return;
    }
 
@@ -1565,6 +1576,7 @@ cmd_calibrate(void)
    real sc, z;
    unsigned long qmask, m;
    int quantity;
+   filepos fp;
 
    qmask = get_qlist(BIT(Q_POS)|BIT(Q_PLUMB)|BIT(Q_LEVEL));
    if (!qmask) return; /* error already reported */
@@ -1579,11 +1591,12 @@ cmd_calibrate(void)
        *
        * *calibrate tape compass 1 1
        */
-      compile_error_skip(/*Can’t calibrate angular and length quantities together*/227);
+      compile_diagnostic(DIAG_ERR|DIAG_SKIP, /*Can’t calibrate angular and length quantities together*/227);
       return;
    }
 
    z = read_numeric(fFalse);
+   get_pos(&fp);
    sc = read_numeric(fTrue);
    if (sc == HUGE_REAL) {
       if (isalpha(ch)) {
@@ -1610,15 +1623,19 @@ cmd_calibrate(void)
    if (sc == HUGE_REAL) sc = (real)1.0;
    /* check for declination scale */
    if (TSTBIT(qmask, Q_DECLINATION) && sc != 1.0) {
+      set_pos(&fp);
       /* TRANSLATORS: DECLINATION is a built-in keyword, so best not to
        * translate */
-      compile_error_skip(-/*Scale factor must be 1.0 for DECLINATION*/40);
+      compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Scale factor must be 1.0 for DECLINATION*/40);
+      skipline();
       return;
    }
    if (sc == 0.0) {
+      set_pos(&fp);
       /* TRANSLATORS: If the scale factor for an instrument is zero, then any
        * reading would be mapped to zero, which doesn't make sense. */
-      compile_error_skip(-/*Scale factor must be non-zero*/391);
+      compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Scale factor must be non-zero*/391);
+      skipline();
       return;
    }
    for (quantity = 0, m = BIT(quantity); m <= qmask; quantity++, m <<= 1) {
@@ -1636,7 +1653,7 @@ cmd_declination(void)
     if (v == HUGE_REAL) {
 	get_token_no_blanks();
 	if (strcmp(ucbuffer, "AUTO") != 0) {
-	    compile_error_skip(-/*Expected number or 'AUTO'*/309);
+	    compile_diagnostic(DIAG_ERR|DIAG_SKIP|DIAG_COL, /*Expected number or 'AUTO'*/309);
 	    return;
 	}
 	/* *declination auto X Y Z */
@@ -1644,7 +1661,7 @@ cmd_declination(void)
 	real y = read_numeric(fFalse);
 	real z = read_numeric(fFalse);
 	if (!pcs->proj) {
-	    compile_error(/*Input coordinate system must be specified for '*DECLINATION AUTO'*/301);
+	    compile_diagnostic(DIAG_ERR, /*Input coordinate system must be specified for '*DECLINATION AUTO'*/301);
 	    return;
 	}
 	if (!proj_wgs84) {
@@ -1658,7 +1675,7 @@ cmd_declination(void)
 	}
 	int r = pj_transform(pcs->proj, proj_wgs84, 1, 1, &x, &y, &z);
 	if (r != 0) {
-	    compile_error(/*Failed to convert coordinates: %s*/436, pj_strerrno(r));
+	    compile_diagnostic(DIAG_ERR, /*Failed to convert coordinates: %s*/436, pj_strerrno(r));
 	    return;
 	}
 	pcs->z[Q_DECLINATION] = HUGE_REAL;
@@ -1699,10 +1716,10 @@ cmd_default(void)
 
    if (default_depr_count < 5) {
       /* TRANSLATORS: If you're unsure what "deprecated" means, see:
-       * http://en.wikipedia.org/wiki/Deprecation */
-      compile_warning(-/**DEFAULT is deprecated - use *CALIBRATE/DATA/SD/UNITS with argument DEFAULT instead*/20);
+       * https://en.wikipedia.org/wiki/Deprecation */
+      compile_diagnostic(DIAG_WARN|DIAG_COL, /**DEFAULT is deprecated - use *CALIBRATE/DATA/SD/UNITS with argument DEFAULT instead*/20);
       if (++default_depr_count == 5)
-	 compile_warning(/*Further uses of this deprecated feature will not be reported*/95);
+	 compile_diagnostic(DIAG_WARN, /*Further uses of this deprecated feature will not be reported*/95);
    }
 
    get_token();
@@ -1718,8 +1735,7 @@ cmd_default(void)
       default_units(pcs);
       break;
     default:
-      file.lpos += strlen(buffer);
-      compile_error_skip(-/*Unknown setting “%s”*/41, buffer);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Unknown setting “%s”*/41, buffer);
    }
 }
 #endif
@@ -1774,7 +1790,7 @@ cmd_sd(void)
    }
    sd = read_numeric(fFalse);
    if (sd <= (real)0.0) {
-      compile_error_skip(-/*Standard deviation must be positive*/48);
+      compile_diagnostic(DIAG_ERR|DIAG_SKIP|DIAG_COL, /*Standard deviation must be positive*/48);
       return;
    }
    units = get_units(qmask, fFalse);
@@ -1820,9 +1836,7 @@ cmd_case(void)
    if (setting != -1) {
       pcs->Case = setting;
    } else {
-      file.lpos += strlen(buffer);
-      compile_error_skip(-/*Found “%s”, expecting “PRESERVE”, “TOUPPER”, or “TOLOWER”*/10,
-			 buffer);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Found “%s”, expecting “PRESERVE”, “TOUPPER”, or “TOLOWER”*/10, buffer);
    }
 }
 
@@ -1873,9 +1887,10 @@ cmd_cs(void)
    if (!had_cs) {
       had_cs = fTrue;
       if (first_fix_name) {
-	 compile_error_at(first_fix_filename, first_fix_line,
-			  /*Station “%s” fixed before CS command first used*/442,
-			  sprint_prefix(first_fix_name));
+	 compile_diagnostic_at(DIAG_ERR,
+			       first_fix_filename, first_fix_line,
+			       /*Station “%s” fixed before CS command first used*/442,
+			       sprint_prefix(first_fix_name));
       }
    }
 
@@ -1885,6 +1900,7 @@ cmd_cs(void)
    get_token();
    if (strcmp(ucbuffer, "OUT") == 0) {
       output = fTrue;
+      get_pos(&fp);
       get_token();
    }
    cs = match_tok(cs_tab, TABSIZE(cs_tab));
@@ -1893,6 +1909,7 @@ cmd_cs(void)
 	 break;
       case CS_CUSTOM:
 	 ok_for_output = MAYBE;
+	 get_pos(&fp);
 	 read_string(&proj_str, &proj_str_len);
 	 cs_sub = 0;
 	 break;
@@ -1989,7 +2006,8 @@ cmd_cs(void)
    }
    if (cs_sub == INT_MIN || isalnum(ch)) {
       set_pos(&fp);
-      compile_error_skip(-/*Unknown coordinate system*/434);
+      compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Unknown coordinate system*/434);
+      skipline();
       return;
    }
    /* Actually handle the cs */
@@ -2064,14 +2082,16 @@ cmd_cs(void)
    if (!proj_str) {
       /* printf("CS %d:%d\n", (int)cs, cs_sub); */
       set_pos(&fp);
-      compile_error_skip(-/*Unknown coordinate system*/434);
+      compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Unknown coordinate system*/434);
+      skipline();
       return;
    }
 
    if (output) {
       if (ok_for_output == NO) {
 	 set_pos(&fp);
-	 compile_error_skip(-/*Coordinate system unsuitable for output*/435);
+	 compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Coordinate system unsuitable for output*/435);
+	 skipline();
 	 return;
       }
 
@@ -2084,12 +2104,15 @@ cmd_cs(void)
 	 projPJ pj = pj_init_plus(proj_str);
 	 if (!pj) {
 	    set_pos(&fp);
-	    compile_error_skip(-/*Invalid coordinate system: %s*/443,
+	    compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Invalid coordinate system: %s*/443,
 			       pj_strerrno(pj_errno));
+	    skipline();
 	    return;
 	 }
 	 if (ok_for_output == MAYBE && pj_is_latlong(pj)) {
-	    compile_error_skip(-/*Coordinate system unsuitable for output*/435);
+	    set_pos(&fp);
+	    compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Coordinate system unsuitable for output*/435);
+	    skipline();
 	    return;
 	 }
 	 if (proj_out) {
@@ -2109,8 +2132,9 @@ cmd_cs(void)
 	 pj = pj_init_plus(proj_str);
 	 if (!pj) {
 	    set_pos(&fp);
-	    compile_error_skip(-/*Invalid coordinate system: %s*/443,
+	    compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Invalid coordinate system: %s*/443,
 			       pj_strerrno(pj_errno));
+	    skipline();
 	    return;
 	 }
       }
@@ -2148,15 +2172,13 @@ cmd_infer(void)
    get_token();
    setting = match_tok(infer_tab, TABSIZE(infer_tab));
    if (setting == INFER_NULL) {
-      file.lpos += strlen(buffer);
-      compile_error_skip(-/*Found “%s”, expecting “EQUATES”, “EXPORTS”, or “PLUMBS”*/31, buffer);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Found “%s”, expecting “EQUATES”, “EXPORTS”, or “PLUMBS”*/31, buffer);
       return;
    }
    get_token();
    on = match_tok(onoff_tab, TABSIZE(onoff_tab));
    if (on == -1) {
-      file.lpos += strlen(buffer);
-      compile_error_skip(-/*Found “%s”, expecting “ON” or “OFF”*/32, buffer);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Found “%s”, expecting “ON” or “OFF”*/32, buffer);
       return;
    }
 
@@ -2266,17 +2288,21 @@ cmd_date(void)
     int year, month, day;
     int days1, days2;
     bool implicit_range = fFalse;
+    filepos fp, fp2;
 
+    get_pos(&fp);
     read_date(&year, &month, &day);
     days1 = days_since_1900(year, month ? month : 1, day ? day : 1);
 
     if (days1 > current_days_since_1900) {
-	compile_warning(-/*Date is in the future!*/80);
+	set_pos(&fp);
+	compile_diagnostic(DIAG_WARN|DIAG_DATE, /*Date is in the future!*/80);
     }
 
     skipblanks();
     if (ch == '-') {
 	nextch();
+	get_pos(&fp2);
 	read_date(&year, &month, &day);
     } else {
 	if (month && day) {
@@ -2291,11 +2317,13 @@ cmd_date(void)
     days2 = days_since_1900(year, month, day);
 
     if (!implicit_range && days2 > current_days_since_1900) {
-	compile_warning(-/*Date is in the future!*/80);
+	set_pos(&fp2);
+	compile_diagnostic(DIAG_WARN|DIAG_DATE, /*Date is in the future!*/80);
     }
 
     if (days2 < days1) {
-	compile_error(-/*End of date range is before the start*/81);
+	set_pos(&fp);
+	compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*End of date range is before the start*/81);
 	int tmp = days1;
 	days1 = days2;
 	days2 = tmp;
@@ -2357,8 +2385,7 @@ handle_command(void)
    cmdtok = match_tok(cmd_tab, TABSIZE(cmd_tab));
 
    if (cmdtok < 0 || cmdtok >= (int)(sizeof(cmd_funcs) / sizeof(cmd_fn))) {
-      file.lpos += strlen(buffer);
-      compile_error_skip(-/*Unknown command “%s”*/12, buffer);
+      compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Unknown command “%s”*/12, buffer);
       return;
    }
 
@@ -2372,7 +2399,7 @@ handle_command(void)
 	  * 1 2 1.23 045 -6
 	  * *export 2
 	  * *end fred */
-	 compile_error(/**EXPORT must immediately follow “*BEGIN <SURVEY>”*/57);
+	 compile_diagnostic(DIAG_ERR, /**EXPORT must immediately follow “*BEGIN <SURVEY>”*/57);
       break;
     case CMD_ALIAS:
     case CMD_CALIBRATE:
