@@ -66,6 +66,8 @@ static real variance[Fr - 1];
 #define VAR(N) variance[(N)-1]
 static long location[Fr - 1];
 #define LOC(N) location[(N)-1]
+static int location_width[Fr - 1];
+#define WID(N) location_width[(N)-1]
 
 /* style functions */
 static void data_normal(void);
@@ -91,13 +93,6 @@ set_pos(const filepos *fp)
       fatalerror_in_file(file.filename, 0, /*Error reading file*/18);
 }
 
-static void
-push_back(int c)
-{
-   if (c != EOF && ungetc(c, file.fh) == EOF)
-      fatalerror_in_file(file.filename, 0, /*Error reading file*/18);
-}
-
 static void
 report_parent(parse * p) {
     if (p->parent)
@@ -124,45 +119,154 @@ error_list_parent_files(void)
 }
 
 static void
-compile_v_report(int severity, int en, va_list ap)
+show_line(int col, int width)
+{
+   /* Rewind to beginning of line. */
+   long cur_pos = ftell(file.fh);
+   int tabs = 0;
+   if (cur_pos < 0 || fseek(file.fh, file.lpos, SEEK_SET) == -1)
+      fatalerror_in_file(file.filename, 0, /*Error reading file*/18);
+
+   /* Read the whole line and write it out. */
+   PUTC(' ', STDERR);
+   while (!feof(file.fh)) {
+      int c = GETC(file.fh);
+      if (isEol(c)) break;
+      if (c == '\t') ++tabs;
+      PUTC(c, STDERR);
+   }
+   fputnl(STDERR);
+
+   /* If we have a location in the line for the error, indicate it. */
+   if (col) {
+      PUTC(' ', STDERR);
+      if (tabs == 0) {
+	 while (--col) PUTC(' ', STDERR);
+      } else {
+	 /* Copy tabs from line, replacing other characters with spaces - this
+	  * means that the caret should line up correctly. */
+	 if (fseek(file.fh, file.lpos, SEEK_SET) == -1)
+	    fatalerror_in_file(file.filename, 0, /*Error reading file*/18);
+	 while (--col) {
+	    int c = GETC(file.fh);
+	    if (c != '\t') c = ' ';
+	    PUTC(c, STDERR);
+	 }
+      }
+      PUTC('^', STDERR);
+      while (width > 1) {
+	 PUTC('~', STDERR);
+	 --width;
+      }
+      fputnl(STDERR);
+   }
+
+   /* Revert to where we were. */
+   if (fseek(file.fh, cur_pos, SEEK_SET) == -1)
+      fatalerror_in_file(file.filename, 0, /*Error reading file*/18);
+}
+
+static int caret_width = 0;
+
+static void
+compile_v_report_fpos(int severity, long fpos, int en, va_list ap)
 {
    int col = 0;
    error_list_parent_files();
-   if (en < 0) {
-      en = -en;
-      if (file.fh) col = ftell(file.fh) - file.lpos;
-   }
+   if (fpos >= file.lpos)
+      col = fpos - file.lpos - caret_width;
    v_report(severity, file.filename, file.line, col, en, ap);
-}
-
-void
-compile_error(int en, ...)
-{
-   va_list ap;
-   va_start(ap, en);
-   compile_v_report(1, en, ap);
-   va_end(ap);
-}
-
-void
-compile_error_skip(int en, ...)
-{
-   va_list ap;
-   va_start(ap, en);
-   compile_v_report(1, en, ap);
-   va_end(ap);
-   skipline();
+   if (file.fh) show_line(col, caret_width);
 }
 
 static void
-compile_error_reading(reading r, int en, ...)
+compile_v_report(int diag_flags, int en, va_list ap)
+{
+   int severity = (diag_flags & DIAG_SEVERITY_MASK);
+   if (diag_flags & (DIAG_COL|DIAG_BUF)) {
+      if (file.fh) {
+	 if (diag_flags & DIAG_BUF) caret_width = strlen(buffer);
+	 compile_v_report_fpos(severity, ftell(file.fh), en, ap);
+	 if (diag_flags & DIAG_BUF) caret_width = 0;
+	 if (diag_flags & DIAG_SKIP) skipline();
+	 return;
+      }
+   }
+   error_list_parent_files();
+   v_report(severity, file.filename, file.line, 0, en, ap);
+   if (file.fh) {
+      if (diag_flags & DIAG_BUF) {
+	 show_line(0, strlen(buffer));
+      } else {
+	 show_line(0, caret_width);
+      }
+   }
+   if (diag_flags & DIAG_SKIP) skipline();
+}
+
+void
+compile_diagnostic(int diag_flags, int en, ...)
 {
    va_list ap;
-   int col = 0;
    va_start(ap, en);
-   error_list_parent_files();
-   if (LOC(r) > file.lpos) col = LOC(r) - file.lpos;
-   v_report(1, file.filename, file.line, col, en, ap);
+   if (diag_flags & (DIAG_TOKEN|DIAG_UINT|DIAG_DATE|DIAG_NUM)) {
+      char *p = NULL;
+      int len = 0;
+      skipblanks();
+      if (diag_flags & DIAG_TOKEN) {
+	 while (!isBlank(ch) && !isEol(ch)) {
+	    s_catchar(&p, &len, (char)ch);
+	    nextch();
+	 }
+      } else if (diag_flags & DIAG_UINT) {
+	 while (isdigit(ch)) {
+	    s_catchar(&p, &len, (char)ch);
+	    nextch();
+	 }
+      } else if (diag_flags & DIAG_DATE) {
+	 while (isdigit(ch) || ch == '.') {
+	    s_catchar(&p, &len, (char)ch);
+	    nextch();
+	 }
+      } else {
+	 if (isMinus(ch) || isPlus(ch)) {
+	    s_catchar(&p, &len, (char)ch);
+	    nextch();
+	 }
+	 while (isdigit(ch)) {
+	    s_catchar(&p, &len, (char)ch);
+	    nextch();
+	 }
+	 if (isDecimal(ch)) {
+	    s_catchar(&p, &len, (char)ch);
+	    nextch();
+	 }
+	 while (isdigit(ch)) {
+	    s_catchar(&p, &len, (char)ch);
+	    nextch();
+	 }
+      }
+      if (p) {
+	 caret_width = strlen(p);
+	 osfree(p);
+      }
+      compile_v_report(diag_flags|DIAG_COL, en, ap);
+      caret_width = 0;
+   } else {
+      compile_v_report(diag_flags, en, ap);
+   }
+   va_end(ap);
+}
+
+static void
+compile_diagnostic_reading(int diag_flags, reading r, int en, ...)
+{
+   va_list ap;
+   int severity = (diag_flags & DIAG_SEVERITY_MASK);
+   va_start(ap, en);
+   caret_width = WID(r);
+   compile_v_report_fpos(severity, LOC(r) + caret_width, en, ap);
+   caret_width = 0;
    va_end(ap);
 }
 
@@ -170,35 +274,36 @@ static void
 compile_error_reading_skip(reading r, int en, ...)
 {
    va_list ap;
-   int col = 0;
    va_start(ap, en);
-   error_list_parent_files();
-   if (LOC(r) > file.lpos) col = LOC(r) - file.lpos;
-   v_report(1, file.filename, file.line, col, en, ap);
+   caret_width = WID(r);
+   compile_v_report_fpos(1, LOC(r) + caret_width, en, ap);
+   caret_width = 0;
    va_end(ap);
    skipline();
 }
 
 void
-compile_error_at(const char * filename, unsigned line, int en, ...)
+compile_diagnostic_at(int diag_flags, const char * filename, unsigned line, int en, ...)
 {
    va_list ap;
+   int severity = (diag_flags & DIAG_SEVERITY_MASK);
    va_start(ap, en);
-   v_report(1, filename, line, 0, en, ap);
+   v_report(severity, filename, line, 0, en, ap);
    va_end(ap);
 }
 
 void
-compile_error_pfx(const prefix * pfx, int en, ...)
+compile_diagnostic_pfx(int diag_flags, const prefix * pfx, int en, ...)
 {
    va_list ap;
+   int severity = (diag_flags & DIAG_SEVERITY_MASK);
    va_start(ap, en);
-   v_report(1, pfx->filename, pfx->line, 0, en, ap);
+   v_report(severity, pfx->filename, pfx->line, 0, en, ap);
    va_end(ap);
 }
 
 void
-compile_error_token(int en)
+compile_diagnostic_token_show(int diag_flags, int en)
 {
    char *p = NULL;
    int len = 0;
@@ -207,35 +312,25 @@ compile_error_token(int en)
       s_catchar(&p, &len, (char)ch);
       nextch();
    }
-   compile_error(en, p ? p : "");
-   osfree(p);
+   if (p) {
+      caret_width = strlen(p);
+      compile_diagnostic(diag_flags|DIAG_COL, en, p);
+      caret_width = 0;
+      osfree(p);
+   } else {
+      compile_diagnostic(DIAG_ERR|DIAG_COL, en, "");
+   }
 }
 
-void
-compile_warning(int en, ...)
+static void
+compile_error_string(const char * s, int en, ...)
 {
    va_list ap;
    va_start(ap, en);
-   compile_v_report(0, en, ap);
-   va_end(ap);
-}
-
-void
-compile_warning_at(const char * filename, unsigned line, int en, ...)
-{
-   va_list ap;
-   va_start(ap, en);
-   v_report(0, filename, line, 0, en, ap);
-   va_end(ap);
-}
-
-void
-compile_warning_pfx(const prefix * pfx, int en, ...)
-{
-   va_list ap;
-   va_start(ap, en);
-   v_report(0, pfx->filename, pfx->line, 0, en, ap);
+   caret_width = strlen(s);
+   compile_v_report(DIAG_ERR|DIAG_COL, en, ap);
    va_end(ap);
+   caret_width = 0;
 }
 
 /* This function makes a note where to put output files */
@@ -275,13 +370,6 @@ skipline(void)
    while (!isEol(ch)) nextch();
 }
 
-static void
-process_bol(void)
-{
-   nextch();
-   skipblanks();
-}
-
 static void
 process_eol(void)
 {
@@ -290,7 +378,8 @@ process_eol(void)
    skipblanks();
 
    if (!isEol(ch)) {
-      if (!isComm(ch)) compile_error(-/*End of line not blank*/15);
+      if (!isComm(ch))
+	 compile_diagnostic(DIAG_ERR|DIAG_COL, /*End of line not blank*/15);
       skipline();
    }
 
@@ -302,18 +391,17 @@ process_eol(void)
    while (ch != EOF) {
       nextch();
       if (ch == eolchar || !isEol(ch)) {
-	 push_back(ch);
 	 break;
       }
       if (ch == '\n') eolchar = ch;
    }
-   file.lpos = ftell(file.fh);
+   file.lpos = ftell(file.fh) - 1;
 }
 
 static bool
 process_non_data_line(void)
 {
-   process_bol();
+   skipblanks();
 
    if (isData(ch)) return fFalse;
 
@@ -354,6 +442,7 @@ read_reading(reading r, bool f_optional)
    }
    LOC(r) = ftell(file.fh);
    VAL(r) = read_numeric_multi(f_optional, &n_readings);
+   WID(r) = ftell(file.fh) - LOC(r);
    VAR(r) = var(q);
    if (n_readings > 1) VAR(r) /= sqrt(n_readings);
 }
@@ -365,6 +454,7 @@ read_bearing_or_omit(reading r)
    q_quantity q = Q_NULL;
    LOC(r) = ftell(file.fh);
    VAL(r) = read_numeric_multi_or_omit(&n_readings);
+   WID(r) = ftell(file.fh) - LOC(r);
    switch (r) {
       case Comp: q = Q_BEARING; break;
       case BackComp: q = Q_BACKBEARING; break;
@@ -383,7 +473,6 @@ nextch_handling_eol(void)
    nextch();
    while (ch != EOF && isEol(ch)) {
       process_eol();
-      nextch();
    }
 }
 
@@ -411,7 +500,7 @@ data_file(const char *pth, const char *fnm)
       }
 
       if (fh == NULL) {
-	 compile_error(-/*Couldn’t open file “%s”*/24, fnm);
+	 compile_error_string(fnm, /*Couldn’t open file “%s”*/24, fnm);
 	 return;
       }
 
@@ -429,6 +518,7 @@ data_file(const char *pth, const char *fnm)
       file.line = 1;
       file.lpos = 0;
       file.reported_where = fFalse;
+      nextch();
    }
 
    using_data_file(file.filename);
@@ -524,7 +614,6 @@ data_file(const char *pth, const char *fnm)
 	    CompassDATFlags, IgnoreAll
 	 };
 	 /* <Cave name> */
-	 process_bol();
 	 skipline();
 	 process_eol();
 	 /* SURVEY NAME: <Short name> */
@@ -597,19 +686,15 @@ data_file(const char *pth, const char *fnm)
 	 skipline();
 	 process_eol();
 	 /* BLANK LINE */
-	 process_bol();
 	 skipline();
 	 process_eol();
 	 /* heading line */
-	 process_bol();
 	 skipline();
 	 process_eol();
 	 /* BLANK LINE */
-	 process_bol();
 	 skipline();
 	 process_eol();
 	 while (!feof(file.fh)) {
-	    process_bol();
 	    if (ch == '\x0c') {
 	       nextch();
 	       process_eol();
@@ -677,9 +762,9 @@ data_file(const char *pth, const char *fnm)
 		     } else {
 			if (x != POS(stn, 0) || y != POS(stn, 1) ||
 			    z != POS(stn, 2)) {
-			   compile_error(/*Station already fixed or equated to a fixed point*/46);
+			   compile_diagnostic(DIAG_ERR, /*Station already fixed or equated to a fixed point*/46);
 			} else {
-			   compile_warning(/*Station already fixed at the same coordinates*/55);
+			   compile_diagnostic(DIAG_WARN, /*Station already fixed at the same coordinates*/55);
 			}
 		     }
 		     while (ch != ']' && ch != EOF) nextch_handling_eol();
@@ -849,7 +934,7 @@ warn_readings_differ(int msgno, real diff, int units)
       }
    }
    strcpy(p, get_units_string(units));
-   compile_warning(msgno, buf);
+   compile_diagnostic(DIAG_WARN, msgno, buf);
 }
 
 static bool
@@ -862,7 +947,7 @@ handle_comp_units(void)
       if (VAL(Comp) < (real)0.0 || VAL(Comp) - M_PI * 2.0 > EPSILON) {
 	 /* TRANSLATORS: Suspicious means something like 410 degrees or -20
 	  * degrees */
-	 compile_warning(/*Suspicious compass reading*/59);
+	 compile_diagnostic_reading(DIAG_WARN, Comp, /*Suspicious compass reading*/59);
 	 VAL(Comp) = mod2pi(VAL(Comp));
       }
    }
@@ -871,7 +956,7 @@ handle_comp_units(void)
       VAL(BackComp) *= pcs->units[Q_BACKBEARING];
       if (VAL(BackComp) < (real)0.0 || VAL(BackComp) - M_PI * 2.0 > EPSILON) {
 	 /* FIXME: different message for BackComp? */
-	 compile_warning(/*Suspicious compass reading*/59);
+	 compile_diagnostic_reading(DIAG_WARN, BackComp, /*Suspicious compass reading*/59);
 	 VAL(BackComp) = mod2pi(VAL(BackComp));
       }
    }
@@ -894,7 +979,7 @@ handle_compass(real *p_var)
       declination = pcs->declination;
    } else {
       if (!pcs->meta || pcs->meta->days1 == -1) {
-	  compile_warning(/*No survey date specified - using 0 for magnetic declination*/304);
+	  compile_diagnostic(DIAG_WARN, /*No survey date specified - using 0 for magnetic declination*/304);
 	  declination = 0;
       } else {
 	  int avg_days = (pcs->meta->days1 + pcs->meta->days2) / 2;
@@ -942,6 +1027,58 @@ handle_compass(real *p_var)
    return comp;
 }
 
+static real
+handle_clino(q_quantity q, reading r, real val, bool percent, clino_type *p_ctype)
+{
+   bool range_0_180;
+   real z;
+   real diff_from_abs90;
+   val *= pcs->units[q];
+   /* percentage scale */
+   if (percent) val = atan(val);
+   /* We want to warn if there's a reading which it would be impossible
+    * to have read from the instrument (e.g. on a -90 to 90 degree scale
+    * you can't read "96" (it's probably a typo for "69").  However, the
+    * gradient reading from a topofil is typically in the range 0 to 180,
+    * with 90 being horizontal.
+    *
+    * Really we should allow the valid range to be specified, but for now
+    * we infer it from the zero error - if this is within 45 degrees of
+    * 90 then we assume the range is 0 to 180.
+    */
+   z = pcs->z[q];
+   range_0_180 = (z > M_PI_4 && z < 3*M_PI_4);
+   diff_from_abs90 = fabs(val) - M_PI_2;
+   if (diff_from_abs90 > EPSILON) {
+      if (!range_0_180) {
+	 int clino_units = get_angle_units(q);
+	 const char * units = get_units_string(clino_units);
+	 real right_angle = M_PI_2 / get_units_factor(clino_units);
+	 /* FIXME: different message for BackClino? */
+	 /* TRANSLATORS: %.f%s will be replaced with a right angle in the
+	  * units currently in use, e.g. "90°" or "100ᵍ".  And "absolute
+	  * value" means the reading ignoring the sign (so it might be
+	  * < -90° or > 90°. */
+	 compile_diagnostic_reading(DIAG_WARN, r, /*Clino reading over %.f%s (absolute value)*/51,
+				    right_angle, units);
+      }
+   } else if (TSTBIT(pcs->infer, INFER_PLUMBS) &&
+	      diff_from_abs90 >= -EPSILON) {
+      *p_ctype = CTYPE_INFERPLUMB;
+   }
+   if (range_0_180 && *p_ctype != CTYPE_INFERPLUMB) {
+      /* FIXME: Warning message not ideal... */
+      if (val < 0.0 || val - M_PI > EPSILON) {
+	 int clino_units = get_angle_units(q);
+	 const char * units = get_units_string(clino_units);
+	 real right_angle = M_PI_2 / get_units_factor(clino_units);
+	 compile_diagnostic_reading(DIAG_WARN, r, /*Clino reading over %.f%s (absolute value)*/51,
+				    right_angle, units);
+      }
+   }
+   return val;
+}
+
 static int
 process_normal(prefix *fr, prefix *to, bool fToFirst,
 	       clino_type ctype, clino_type backctype)
@@ -961,78 +1098,19 @@ process_normal(prefix *fr, prefix *to, bool fToFirst,
    /* adjusted tape is negative -- probably the calibration is wrong */
    if (tape < (real)0.0) {
       /* TRANSLATE different message for topofil? */
-      compile_warning(/*Negative adjusted tape reading*/79);
+      compile_diagnostic_reading(DIAG_WARN, Tape, /*Negative adjusted tape reading*/79);
    }
 
    fNoComp = handle_comp_units();
 
    if (ctype == CTYPE_READING) {
-      bool range_0_180;
-      real z;
-      real diff_from_abs90;
-      clin *= pcs->units[Q_GRADIENT];
-      /* percentage scale */
-      if (pcs->f_clino_percent) clin = atan(clin);
-      /* We want to warn if there's a reading which it would be impossible
-       * to have read from the instrument (e.g. on a -90 to 90 degree scale
-       * you can't read "96" (it's probably a typo for "69").  However, the
-       * gradient reading from a topofil is typically in the range 0 to 180,
-       * with 90 being horizontal.
-       *
-       * Really we should allow the valid range to be specified, but for now
-       * we infer it from the zero error - if this is within 45 degrees of
-       * 90 then we assume the range is 0 to 180.
-       */
-      z = pcs->z[Q_GRADIENT];
-      range_0_180 = (z > M_PI_4 && z < 3*M_PI_4);
-      diff_from_abs90 = fabs(clin) - M_PI_2;
-      if (diff_from_abs90 > EPSILON) {
-	 if (!range_0_180) {
-	    int clino_units = get_angle_units(Q_GRADIENT);
-	    const char * units = get_units_string(clino_units);
-	    real right_angle = M_PI_2 / get_units_factor(clino_units);
-	    /* TRANSLATORS: %.f%s will be replaced with a right angle in the
-	     * units currently in use, e.g. "90°" or "100ᵍ".  And "absolute
-	     * value" means the reading ignoring the sign (so it might be
-	     * < -90° or > 90°. */
-	    compile_warning(/*Clino reading over %.f%s (absolute value)*/51,
-			    right_angle, units);
-	 }
-      } else if (TSTBIT(pcs->infer, INFER_PLUMBS) &&
-		 diff_from_abs90 >= -EPSILON) {
-	 ctype = CTYPE_INFERPLUMB;
-      }
-      if (range_0_180 && ctype != CTYPE_INFERPLUMB) {
-	 /* FIXME: Warning message not ideal... */
-	 if (clin < 0.0 || clin - M_PI > EPSILON) {
-	    int clino_units = get_angle_units(Q_GRADIENT);
-	    const char * units = get_units_string(clino_units);
-	    real right_angle = M_PI_2 / get_units_factor(clino_units);
-	    compile_warning(/*Clino reading over %.f%s (absolute value)*/51,
-			    right_angle, units);
-	 }
-      }
+      clin = handle_clino(Q_GRADIENT, Clino, clin,
+			  pcs->f_clino_percent, &ctype);
    }
 
    if (backctype == CTYPE_READING) {
-      backclin *= pcs->units[Q_BACKGRADIENT];
-      /* percentage scale */
-      if (pcs->f_backclino_percent) backclin = atan(backclin);
-      /* FIXME: Add range_0_180 handling here too */
-      if (ctype != CTYPE_READING) {
-	 real diff_from_abs90 = fabs(backclin) - M_PI_2;
-	 if (diff_from_abs90 > EPSILON) {
-	    /* FIXME: different message for BackClino? */
-	    int clino_units = get_angle_units(Q_BACKGRADIENT);
-	    const char * units = get_units_string(clino_units);
-	    real right_angle = M_PI_2 / get_units_factor(clino_units);
-	    compile_warning(/*Clino reading over %.f%s (absolute value)*/51,
-			    right_angle, units);
-	 } else if (TSTBIT(pcs->infer, INFER_PLUMBS) &&
-		    diff_from_abs90 >= -EPSILON) {
-	    backctype = CTYPE_INFERPLUMB;
-	 }
-      }
+      backclin = handle_clino(Q_BACKGRADIENT, BackClino, backclin,
+			      pcs->f_backclino_percent, &backctype);
    }
 
    /* un-infer the plumb if the backsight was just a reading */
@@ -1060,7 +1138,7 @@ process_normal(prefix *fr, prefix *to, bool fToFirst,
 	    /* TRANSLATORS: A "plumbed leg" is one measured using a plumbline
 	     * (a weight on a string).  So the problem here is that the leg is
 	     * vertical, so a compass reading has no meaning! */
-	    compile_warning(/*Compass reading given on plumbed leg*/21);
+	    compile_diagnostic(DIAG_WARN, /*Compass reading given on plumbed leg*/21);
 	 }
       }
 
@@ -1234,7 +1312,7 @@ process_diving(prefix *fr, prefix *to, bool fToFirst, bool fDepthChange)
 
    /* adjusted tape is negative -- probably the calibration is wrong */
    if (tape < (real)0.0) {
-      compile_warning(/*Negative adjusted tape reading*/79);
+      compile_diagnostic(DIAG_WARN, /*Negative adjusted tape reading*/79);
    }
 
    /* check if tape is less than depth change */
@@ -1245,7 +1323,7 @@ process_diving(prefix *fr, prefix *to, bool fToFirst, bool fDepthChange)
        * It could be a gross error (e.g. the decimal point is missing from the
        * depth gauge reading) or it could just be due to random error on a near
        * vertical leg */
-      compile_warning(/*Tape reading is less than change in depth*/62);
+      compile_diagnostic(DIAG_WARN, /*Tape reading is less than change in depth*/62);
    }
 
    if (tape == (real)0.0 && dz == 0.0) {
@@ -1379,10 +1457,9 @@ data_cartesian(void)
 	 fMulti = fTrue;
 	 while (1) {
 	    process_eol();
-	    process_bol();
+	    skipblanks();
 	    if (isData(ch)) break;
 	    if (!isComm(ch)) {
-	       push_back(ch);
 	       return;
 	    }
 	 }
@@ -1398,7 +1475,7 @@ data_cartesian(void)
 	 }
 	 do {
 	    process_eol();
-	    process_bol();
+	    skipblanks();
 	 } while (isComm(ch));
 	 goto again;
        default: BUG("Unknown reading in ordering");
@@ -1431,7 +1508,7 @@ process_cylpolar(prefix *fr, prefix *to, bool fToFirst, bool fDepthChange)
 
    /* adjusted tape is negative -- probably the calibration is wrong */
    if (tape < (real)0.0) {
-      compile_warning(/*Negative adjusted tape reading*/79);
+      compile_diagnostic(DIAG_WARN, /*Negative adjusted tape reading*/79);
    }
 
    if (VAL(Comp) == HUGE_REAL && VAL(BackComp) == HUGE_REAL) {
@@ -1539,9 +1616,7 @@ data_normal(void)
 	     fRev = fTrue;
 	     break;
 	   default:
-	     file.lpos += strlen(buffer);
-	     compile_error_skip(-/*Found “%s”, expecting “F” or “B”*/131,
-				buffer);
+	     compile_diagnostic(DIAG_ERR|DIAG_BUF|DIAG_SKIP, /*Found “%s”, expecting “F” or “B”*/131, buffer);
 	     process_eol();
 	     return;
 	  }
@@ -1552,20 +1627,21 @@ data_normal(void)
 	  read_reading(r, fTrue);
 	  if (VAL(r) == HUGE_REAL) {
 	     if (!isOmit(ch)) {
-		compile_error_token(-/*Expecting numeric field, found “%s”*/9);
+		compile_diagnostic_token_show(DIAG_ERR, /*Expecting numeric field, found “%s”*/9);
 		/* Avoid also warning about omitted tape reading. */
 		VAL(r) = 0;
 	     } else {
 		nextch();
 	     }
 	  } else if (VAL(r) < (real)0.0) {
-	     compile_warning(-/*Negative tape reading*/60);
+	     compile_diagnostic_reading(DIAG_WARN, r, /*Negative tape reading*/60);
 	  }
 	  break;
        }
        case Count:
 	  VAL(FrCount) = VAL(ToCount);
 	  LOC(FrCount) = LOC(ToCount);
+	  WID(FrCount) = WID(ToCount);
 	  read_reading(ToCount, fFalse);
 	  fTopofil = fTrue;
 	  break;
@@ -1586,7 +1662,7 @@ data_normal(void)
 	  if (VAL(r) == HUGE_REAL) {
 	     VAL(r) = handle_plumb(p_ctype);
 	     if (VAL(r) != HUGE_REAL) break;
-	     compile_error_token(-/*Expecting numeric field, found “%s”*/9);
+	     compile_diagnostic_token_show(DIAG_ERR, /*Expecting numeric field, found “%s”*/9);
 	     skipline();
 	     process_eol();
 	     return;
@@ -1600,6 +1676,7 @@ data_normal(void)
        case Depth:
 	  VAL(FrDepth) = VAL(ToDepth);
 	  LOC(FrDepth) = LOC(ToDepth);
+	  WID(FrDepth) = WID(ToDepth);
 	  read_reading(ToDepth, fFalse);
 	  break;
        case DepthChange:
@@ -1644,10 +1721,10 @@ data_normal(void)
        }
        case CompassDATFlags:
 	  if (ch == '#') {
+	     filepos fp;
+	     get_pos(&fp);
 	     nextch();
 	     if (ch == '|') {
-		filepos fp;
-		get_pos(&fp);
 		nextch();
 		while (ch >= 'A' && ch <= 'Z') {
 		   compass_dat_flags |= BIT(ch - 'A');
@@ -1665,12 +1742,9 @@ data_normal(void)
 		} else {
 		   compass_dat_flags = 0;
 		   set_pos(&fp);
-		   push_back('|');
-		   ch = '#';
 		}
 	     } else {
-		push_back(ch);
-		ch = '#';
+		set_pos(&fp);
 	     }
 	  }
 	  break;
@@ -1687,6 +1761,7 @@ data_normal(void)
 	     if (fTopofil) {
 		VAL(Tape) = VAL(ToCount) - VAL(FrCount);
 		LOC(Tape) = LOC(ToCount);
+		WID(Tape) = WID(ToCount);
 	     }
 	     /* Note: frdepth == todepth test works regardless of fDepthChange
 	      * (frdepth always zero, todepth is change of depth) and also
@@ -1731,7 +1806,7 @@ data_normal(void)
 		   VAR(Tape) = VAR(BackTape);
 		}
 	     } else if (VAL(Tape) == HUGE_REAL) {
-		compile_error_reading(Tape, /*Tape reading may not be omitted*/94);
+		compile_diagnostic_reading(DIAG_ERR, Tape, /*Tape reading may not be omitted*/94);
 		goto inferred_equate;
 	     }
 	     implicit_splay = TSTBIT(pcs->flags, FLAGS_IMPLICIT_SPLAY);
@@ -1777,16 +1852,16 @@ data_normal(void)
 	  /* this is also used if clino reading is the omit character */
 	  VAL(Clino) = VAL(BackClino) = 0;
 	  LOC(Clino) = LOC(BackClino) = -1;
+	  WID(Clino) = WID(BackClino) = 0;
 
 	  inferred_equate:
 
 	  fMulti = fTrue;
 	  while (1) {
 	      process_eol();
-	      process_bol();
+	      skipblanks();
 	      if (isData(ch)) break;
 	      if (!isComm(ch)) {
-		 push_back(ch);
 		 return;
 	      }
 	  }
@@ -1811,6 +1886,7 @@ data_normal(void)
 	     if (fTopofil) {
 		VAL(Tape) = VAL(ToCount) - VAL(FrCount);
 		LOC(Tape) = LOC(ToCount);
+		WID(Tape) = WID(ToCount);
 	     }
 	     /* Note: frdepth == todepth test works regardless of fDepthChange
 	      * (frdepth always zero, todepth is change of depth) and also
@@ -1851,7 +1927,7 @@ data_normal(void)
 		   VAR(Tape) = VAR(BackTape);
 		}
 	     } else if (VAL(Tape) == HUGE_REAL) {
-		compile_error_reading(Tape, /*Tape reading may not be omitted*/94);
+		compile_diagnostic_reading(DIAG_ERR, Tape, /*Tape reading may not be omitted*/94);
 		process_eol();
 		return;
 	     }
@@ -1890,7 +1966,7 @@ data_normal(void)
 	  }
 	  do {
 	     process_eol();
-	     process_bol();
+	     skipblanks();
 	  } while (isComm(ch));
 	  goto again;
        default:
@@ -1935,7 +2011,7 @@ data_passage(void)
 	 read_reading(r, fTrue);
 	 if (VAL(r) == HUGE_REAL) {
 	    if (!isOmit(ch)) {
-	       compile_error_token(-/*Expecting numeric field, found “%s”*/9);
+	       compile_diagnostic_token_show(DIAG_ERR, /*Expecting numeric field, found “%s”*/9);
 	    } else {
 	       nextch();
 	    }
@@ -2026,10 +2102,9 @@ data_nosurvey(void)
 	 if (ordering[1] == End) {
 	    do {
 	       process_eol();
-	       process_bol();
+	       skipblanks();
 	    } while (isComm(ch));
 	    if (!isData(ch)) {
-	       push_back(ch);
 	       return;
 	    }
 	    goto again;
@@ -2037,10 +2112,9 @@ data_nosurvey(void)
 	 fMulti = fTrue;
 	 while (1) {
 	    process_eol();
-	    process_bol();
+	    skipblanks();
 	    if (isData(ch)) break;
 	    if (!isComm(ch)) {
-	       push_back(ch);
 	       return;
 	    }
 	 }
@@ -2056,7 +2130,7 @@ data_nosurvey(void)
 	 }
 	 do {
 	    process_eol();
-	    process_bol();
+	    skipblanks();
 	 } while (isComm(ch));
 	 goto again;
        default: BUG("Unknown reading in ordering");
