@@ -75,6 +75,26 @@ using namespace std;
 
 const int BLOB_DIAMETER = 5;
 
+#define BLOB_TEXTURE \
+	    o, o, o, o, o, o, o, o,\
+	    o, o, o, o, o, o, o, o,\
+	    o, o, I, I, I, o, o, o,\
+	    o, I, I, I, I, I, o, o,\
+	    o, I, I, I, I, I, o, o,\
+	    o, I, I, I, I, I, o, o,\
+	    o, o, I, I, I, o, o, o,\
+	    o, o, o, o, o, o, o, o
+
+#define CROSS_TEXTURE \
+	    o, o, o, o, o, o, o, o,\
+	    I, o, o, o, o, o, I, o,\
+	    o, I, o, o, o, I, o, o,\
+	    o, o, I, o, I, o, o, o,\
+	    o, o, o, I, o, o, o, o,\
+	    o, o, I, o, I, o, o, o,\
+	    o, I, o, o, o, I, o, o,\
+	    I, o, o, o, o, o, I, o
+
 static bool opengl_initialised = false;
 
 string GetGLSystemDescription()
@@ -140,6 +160,53 @@ string GetGLSystemDescription()
     return info;
 }
 
+static bool
+glpoint_sprite_works()
+{
+    // Point sprites provide an easy, fast way for us to draw crosses by
+    // texture mapping GL points.
+    //
+    // If we have OpenGL >= 2.0 then we definitely have GL_POINT_SPRITE.
+    // Otherwise see if we have the GL_ARB_point_sprite or GL_NV_point_sprite
+    // extensions.
+    //
+    // The symbolic constants GL_POINT_SPRITE, GL_POINT_SPRITE_ARB, and
+    // GL_POINT_SPRITE_NV all give the same number so it doesn't matter
+    // which we use.
+    static bool glpoint_sprite = false;
+    static bool checked = false;
+    if (!checked) {
+	float maxSize = 0.0f;
+	glGetFloatv(GL_POINT_SIZE_MAX, &maxSize);
+	if (maxSize >= 8) {
+	    glpoint_sprite = (atoi((const char *)glGetString(GL_VERSION)) >= 2);
+	    if (!glpoint_sprite) {
+		const char * p = (const char *)glGetString(GL_EXTENSIONS);
+		while (true) {
+		    size_t l = 0;
+		    if (memcmp(p, "GL_ARB_point_sprite", 19) == 0) {
+			l = 19;
+		    } else if (memcmp(p, "GL_NV_point_sprite", 18) == 0) {
+			l = 18;
+		    }
+		    if (l) {
+			p += l;
+			if (*p == '\0' || *p == ' ') {
+			    glpoint_sprite = true;
+			    break;
+			}
+		    }
+		    p = strchr(p + 1, ' ');
+		    if (!p) break;
+		    ++p;
+		}
+	    }
+	}
+	checked = true;
+    }
+    return glpoint_sprite;
+}
+
 static void
 log_gl_error(const wxChar * str, GLenum error_code)
 {
@@ -247,7 +314,7 @@ bool GLAList::need_to_generate() {
 	// probably a infinite loop in the lazy list mechanism.
 	assert(gl_list < 1000);
     }
-    // http://www.opengl.org/resources/faq/technical/displaylist.htm advises:
+    // https://www.opengl.org/resources/faq/technical/displaylist.htm advises:
     //
     // "Stay away from GL_COMPILE_AND_EXECUTE mode. Instead, create the
     // list using GL_COMPILE mode, then execute it with glCallList()."
@@ -346,15 +413,17 @@ void GLACanvas::FirstShow()
     CHECK_GL_ERROR("FirstShow", "glViewport");
 
     save_hints = false;
+
     vendor = wxString((const char *)glGetString(GL_VENDOR), wxConvUTF8);
     renderer = wxString((const char *)glGetString(GL_RENDERER), wxConvUTF8);
-    wxConfigBase * cfg = wxConfigBase::Get();
     {
+	wxConfigBase * cfg = wxConfigBase::Get();
 	wxString s;
-	if (cfg->Read(wxT("opengl_vendor"), &s, wxString()) && s == vendor &&
+	if (cfg->Read(wxT("opengl_survex"), &s, wxString()) && s == wxT(VERSION) &&
+	    cfg->Read(wxT("opengl_vendor"), &s, wxString()) && s == vendor &&
 	    cfg->Read(wxT("opengl_renderer"), &s, wxString()) && s == renderer) {
-	    // The vendor and renderer are the same as the values we have cached,
-	    // so use the hints we have cached.
+	    // The survex version, vendor and renderer are the same as those
+	    // we cached hints for, so use those hints.
 	    int v;
 	    if (cfg->Read(wxT("blob_method"), &v, 0) &&
 		(v == POINT || v == LINES)) {
@@ -432,7 +501,7 @@ void GLACanvas::FirstShow()
 	    point_size_range[1] >= BLOB_DIAMETER) {
 	    blob_method = POINT;
 	} else {
-	    blob_method = LINES;
+	    blob_method = glpoint_sprite_works() ? SPRITE : LINES;
 	}
 	save_hints = true;
     }
@@ -442,45 +511,8 @@ void GLACanvas::FirstShow()
 	CHECK_GL_ERROR("FirstShow", "glPointSize");
     }
 
-    // Point sprites provide an easy, fast way for us to draw crosses by
-    // texture mapping GL points.
-    //
-    // If we have OpenGL >= 2.0 then we definitely have GL_POINT_SPRITE.
-    // Otherwise see if we have the GL_ARB_point_sprite or GL_NV_point_sprite
-    // extensions.
-    //
-    // The symbolic constants GL_POINT_SPRITE, GL_POINT_SPRITE_ARB, and
-    // GL_POINT_SPRITE_NV all give the same number so it doesn't matter
-    // which we use.
     if (cross_method == UNKNOWN) {
-	bool glpoint_sprite = false;
-	float maxSize = 0.0f;
-	glGetFloatv(GL_POINT_SIZE_MAX, &maxSize);
-	if (maxSize >= 8) {
-	    glpoint_sprite = (atoi((const char *)glGetString(GL_VERSION)) >= 2);
-	    if (!glpoint_sprite) {
-		const char * p = (const char *)glGetString(GL_EXTENSIONS);
-		while (true) {
-		    size_t l = 0;
-		    if (memcmp(p, "GL_ARB_point_sprite", 19) == 0) {
-			l = 19;
-		    } else if (memcmp(p, "GL_NV_point_sprite", 18) == 0) {
-			l = 18;
-		    }
-		    if (l) {
-			p += l;
-			if (*p == '\0' || *p == ' ') {
-			    glpoint_sprite = true;
-			    break;
-			}
-		    }
-		    p = strchr(p + 1, ' ');
-		    if (!p) break;
-		    ++p;
-		}
-	    }
-	}
-	cross_method = glpoint_sprite ? SPRITE : LINES;
+	cross_method = glpoint_sprite_works() ? SPRITE : LINES;
 	save_hints = true;
     }
 
@@ -491,14 +523,11 @@ void GLACanvas::FirstShow()
 	CHECK_GL_ERROR("FirstShow", "glBindTexture");
 	// Cross image for drawing crosses using texture mapped point sprites.
 	const unsigned char crossteximage[128] = {
-	      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	    255,255,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,255,255,  0,  0,
-	      0,  0,255,255,  0,  0,  0,  0,  0,  0,255,255,  0,  0,  0,  0,
-	      0,  0,  0,  0,255,255,  0,  0,255,255,  0,  0,  0,  0,  0,  0,
-	      0,  0,  0,  0,  0,  0,255,255,  0,  0,  0,  0,  0,  0,  0,  0,
-	      0,  0,  0,  0,255,255,  0,  0,255,255,  0,  0,  0,  0,  0,  0,
-	      0,  0,255,255,  0,  0,  0,  0,  0,  0,255,255,  0,  0,  0,  0,
-	    255,255,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,255,255,  0,  0
+#define o 0,0
+#define I 255,255
+	    CROSS_TEXTURE
+#undef o
+#undef I
 	};
 	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 	CHECK_GL_ERROR("FirstShow", "glPixelStorei");
@@ -515,6 +544,35 @@ void GLACanvas::FirstShow()
 	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
 	CHECK_GL_ERROR("FirstShow", "glTexParameteri GL_TEXTURE_MIN_FILTER");
     }
+
+    if (blob_method == SPRITE) {
+	glGenTextures(1, &m_BlobTexture);
+	CHECK_GL_ERROR("FirstShow", "glGenTextures");
+	glBindTexture(GL_TEXTURE_2D, m_BlobTexture);
+	CHECK_GL_ERROR("FirstShow", "glBindTexture");
+	// Image for drawing blobs using texture mapped point sprites.
+	const unsigned char blobteximage[128] = {
+#define o 0,0
+#define I 255,255
+	    BLOB_TEXTURE
+#undef o
+#undef I
+	};
+	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
+	CHECK_GL_ERROR("FirstShow", "glPixelStorei");
+	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
+	CHECK_GL_ERROR("FirstShow", "glTexEnvi");
+	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
+	CHECK_GL_ERROR("FirstShow", "glTexParameteri GL_TEXTURE_WRAP_S");
+	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
+	CHECK_GL_ERROR("FirstShow", "glTexParameteri GL_TEXTURE_WRAP_T");
+	glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE_ALPHA, 8, 8, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, (GLvoid *)blobteximage);
+	CHECK_GL_ERROR("FirstShow", "glTexImage2D");
+	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
+	CHECK_GL_ERROR("FirstShow", "glTexParameteri GL_TEXTURE_MAG_FILTER");
+	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
+	CHECK_GL_ERROR("FirstShow", "glTexParameteri GL_TEXTURE_MIN_FILTER");
+    }
 }
 
 void GLACanvas::Clear()
@@ -618,14 +676,7 @@ void GLACanvas::StartDrawing()
 	static const unsigned char expected_cross[64 * 3] = {
 #define o 0,0,0
 #define I 255,255,255
-	    o, o, o, o, o, o, o, o,
-	    I, o, o, o, o, o, I, o,
-	    o, I, o, o, o, I, o, o,
-	    o, o, I, o, I, o, o, o,
-	    o, o, o, I, o, o, o, o,
-	    o, o, I, o, I, o, o, o,
-	    o, I, o, o, o, I, o, o,
-	    I, o, o, o, o, o, I, o
+	    CROSS_TEXTURE
 #undef o
 #undef I
 	};
@@ -645,14 +696,7 @@ void GLACanvas::StartDrawing()
 	static const unsigned char expected_blob[64 * 3] = {
 #define o 0,0,0
 #define I 255,255,255
-	    o, o, o, o, o, o, o, o,
-	    o, o, o, o, o, o, o, o,
-	    o, o, I, I, I, o, o, o,
-	    o, I, I, I, I, I, o, o,
-	    o, I, I, I, I, I, o, o,
-	    o, I, I, I, I, I, o, o,
-	    o, o, I, I, I, o, o, o,
-	    o, o, o, o, o, o, o, o
+	    BLOB_TEXTURE
 #undef o
 #undef I
 	};
@@ -663,6 +707,7 @@ void GLACanvas::StartDrawing()
     }
 
     wxConfigBase * cfg = wxConfigBase::Get();
+    cfg->Write(wxT("opengl_survex"), wxT(VERSION));
     cfg->Write(wxT("opengl_vendor"), vendor);
     cfg->Write(wxT("opengl_renderer"), renderer);
     cfg->Write(wxT("blob_method"), blob_method);
@@ -1128,7 +1173,23 @@ void GLACanvas::PlaceIndicatorVertex(glaCoord x, glaCoord y)
 void GLACanvas::BeginBlobs()
 {
     // Commence drawing of a set of blobs.
-    if (blob_method == POINT) {
+    if (blob_method == SPRITE) {
+	glPushAttrib(GL_ENABLE_BIT|GL_POINT_BIT);
+	CHECK_GL_ERROR("BeginBlobs", "glPushAttrib");
+	glBindTexture(GL_TEXTURE_2D, m_BlobTexture);
+	CHECK_GL_ERROR("BeginBlobs", "glBindTexture");
+	glEnable(GL_ALPHA_TEST);
+	CHECK_GL_ERROR("BeginBlobs", "glEnable GL_ALPHA_TEST");
+	glPointSize(8);
+	CHECK_GL_ERROR("BeginBlobs", "glPointSize");
+	glTexEnvi(GL_POINT_SPRITE, GL_COORD_REPLACE, GL_TRUE);
+	CHECK_GL_ERROR("BeginBlobs", "glTexEnvi GL_POINT_SPRITE");
+	glEnable(GL_TEXTURE_2D);
+	CHECK_GL_ERROR("BeginBlobs", "glEnable GL_TEXTURE_2D");
+	glEnable(GL_POINT_SPRITE);
+	CHECK_GL_ERROR("BeginBlobs", "glEnable GL_POINT_SPRITE");
+	glBegin(GL_POINTS);
+    } else if (blob_method == POINT) {
 	glPushAttrib(GL_ENABLE_BIT);
 	CHECK_GL_ERROR("BeginBlobs", "glPushAttrib");
 	glEnable(GL_ALPHA_TEST);
@@ -1150,7 +1211,7 @@ void GLACanvas::EndBlobs()
 {
     // Finish drawing of a set of blobs.
     glEnd();
-    if (blob_method == POINT) {
+    if (blob_method != LINES) {
 	CHECK_GL_ERROR("EndBlobs", "glEnd GL_POINTS");
     } else {
 	CHECK_GL_ERROR("EndBlobs", "glEnd GL_LINES");
@@ -1161,7 +1222,7 @@ void GLACanvas::EndBlobs()
 
 void GLACanvas::DrawBlob(glaCoord x, glaCoord y, glaCoord z)
 {
-    if (blob_method == POINT) {
+    if (blob_method != LINES) {
 	// Draw a marker.
 	PlaceVertex(x, y, z);
     } else {
@@ -1195,7 +1256,7 @@ void GLACanvas::DrawBlob(glaCoord x, glaCoord y, glaCoord z)
 
 void GLACanvas::DrawBlob(glaCoord x, glaCoord y)
 {
-    if (blob_method == POINT) {
+    if (blob_method != LINES) {
 	// Draw a marker.
 	PlaceVertex(x, y, 0);
     } else {
@@ -1544,7 +1605,7 @@ bool GLACanvas::CheckVisualFidelity(const unsigned char * target) const
 	glReadBuffer(GL_BACK);
 	CHECK_GL_ERROR("FirstShow", "glReadBuffer");
     }
-    glReadPixels(x_size / 2 - 3, y_size / 2 - 4, 8, 8,
+    glReadPixels(x_size / 2 - 4, y_size / 2 - 5, 8, 8,
 		 GL_RGB, GL_UNSIGNED_BYTE, (GLvoid *)pixels);
     CHECK_GL_ERROR("CheckVisualFidelity", "glReadPixels");
     if (double_buffered) {
