@@ -63,7 +63,7 @@ html_escape(FILE *fh, const char *s)
 }
 
 KML::KML(const char * input_datum)
-    : pj_input(NULL), pj_output(NULL), in_linestring(false)
+    : pj_input(NULL), pj_output(NULL), in_linestring(false), in_wall(false)
 {
     if (!(pj_input = pj_init_plus(input_datum))) {
 	wxString m = wmsg(/*Failed to initialise input coordinate system “%s”*/287);
@@ -88,7 +88,9 @@ KML::~KML()
 const int *
 KML::passes() const
 {
-    static const int default_passes[] = { LABELS|ENTS|FIXES|EXPORTS, LEGS|SURF, 0 };
+    static const int default_passes[] = {
+	PASG, XSECT, WALL1, WALL2, LEGS|SURF, LABELS|ENTS|FIXES|EXPORTS, 0
+    };
     return default_passes;
 }
 
@@ -143,6 +145,103 @@ KML::line(const img_point *p1, const img_point *p, unsigned /*flags*/, bool fPen
     fprintf(fh, "%.8f,%.8f,%.2f\n", X, Y, Z);
 }
 
+void
+KML::xsect(const img_point *p, double angle, double d1, double d2)
+{
+    double s = sin(rad(angle));
+    double c = cos(rad(angle));
+
+    double x1 = p->x + c * d1;
+    double y1 = p->y + s * d1;
+    double z1 = p->z;
+    pj_transform(pj_input, pj_output, 1, 1, &x1, &y1, &z1);
+    x1 = deg(x1);
+    y1 = deg(y1);
+
+    double x2 = p->x - c * d2;
+    double y2 = p->y - s * d2;
+    double z2 = p->z;
+    pj_transform(pj_input, pj_output, 1, 1, &x2, &y2, &z2);
+    x2 = deg(x2);
+    y2 = deg(y2);
+
+    fputs("<Placemark><name></name><LineString><altitudeMode>absolute</altitudeMode><coordinates>", fh);
+    fprintf(fh, "%.8f,%.8f,%.8f %.8f,%.8f,%.8f", x1, y1, z1, x2, y2, z2);
+    fputs("</coordinates></LineString></Placemark>\n", fh);
+}
+
+void
+KML::wall(const img_point *p, double angle, double d)
+{
+    double s = sin(rad(angle));
+    double c = cos(rad(angle));
+
+    double x = p->x + c * d;
+    double y = p->y + s * d;
+    double z = p->z;
+    pj_transform(pj_input, pj_output, 1, 1, &x, &y, &z);
+    x = deg(x);
+    y = deg(y);
+
+    if (!in_wall) {
+	fputs("<Placemark><name></name><LineString><altitudeMode>absolute</altitudeMode><coordinates>", fh);
+	in_wall = true;
+    }
+    fprintf(fh, "%.8f,%.8f,%.8f\n", x, y, z);
+}
+
+void
+KML::passage(const img_point *p, double angle, double d1, double d2)
+{
+    double s = sin(rad(angle));
+    double c = cos(rad(angle));
+
+    // Draw along one side and push the other onto a stack, then at the end pop
+    // the stack and write out those points to give one polygon, fewer points,
+    // and a smaller file.
+    double x1 = p->x + c * d1;
+    double y1 = p->y + s * d1;
+    double z1 = p->z;
+    pj_transform(pj_input, pj_output, 1, 1, &x1, &y1, &z1);
+    x1 = deg(x1);
+    y1 = deg(y1);
+
+    double x2 = p->x - c * d2;
+    double y2 = p->y - s * d2;
+    double z2 = p->z;
+    pj_transform(pj_input, pj_output, 1, 1, &x2, &y2, &z2);
+    x2 = deg(x2);
+    y2 = deg(y2);
+
+    if (psg.empty()) {
+	fprintf(fh, "<Placemark><name></name><Polygon><altitudeMode>absolute</altitudeMode>"
+		    "<outerBoundaryIs><LinearRing><coordinates>\n");
+    }
+    // NB - order of vertices should be anti-clockwise in a KML file, so go
+    // along the right wall now, and put the left wall points on a stack to
+    // come back along at the end.
+    fprintf(fh, "%.8f,%.8f,%.8f\n", x2, y2, z2);
+    psg.push_back(Vector3(x1, y1, z1));
+}
+
+void
+KML::tube_end()
+{
+    if (!psg.empty()) {
+	vector<Vector3>::const_reverse_iterator i;
+	for (i = psg.rbegin(); i != psg.rend(); ++i) {
+	    fprintf(fh, "%.8f,%.8f,%.8f\n", i->GetX(), i->GetY(), i->GetZ());
+	}
+	psg.clear();
+	fputs("</coordinates></LinearRing></outerBoundaryIs>"
+	      "</Polygon></Placemark>\n", fh);
+    }
+    if (in_wall) {
+	fputs("</coordinates></LineString></Placemark>\n", fh);
+	in_wall = false;
+    }
+}
+
 void
 KML::label(const img_point *p, const char *s, bool /*fSurface*/, int type)
 {
