@@ -2579,8 +2579,8 @@ void MainFrm::OnPresExportMovie(wxCommandEvent&)
     wxString baseleaf;
     wxFileName::SplitPath(m_File, NULL, NULL, &baseleaf, NULL, wxPATH_NATIVE);
     wxFileDialog dlg(this, wmsg(/*Export Movie*/331), wxString(),
-		     baseleaf + wxT(".mpg"),
-		     wxT("MPEG|*.mpg|AVI|*.avi|QuickTime|*.mov|WMV|*.wmv;*.asf"),
+		     baseleaf + wxT(".mp4"),
+		     wxT("MPEG|*.mp4|OGG|*.ogv|AVI|*.avi|QuickTime|*.mov|WMV|*.wmv;*.asf"),
 		     wxFD_SAVE|wxFD_OVERWRITE_PROMPT);
     if (dlg.ShowModal() == wxID_OK) {
 	// Error is reported by GfxCore.
@@ -2610,9 +2610,8 @@ void MainFrm::RestrictTo(const wxString & survey)
 	}
 	new_prefix += survey;
     }
-    // FIXME: Be more efficient to reload the processed data rather rather than
-    // potentially reprocessing.
-    if (!LoadData(m_File, new_prefix))
+    // Reload the processed data rather rather than potentially reprocessing.
+    if (!LoadData(m_FileProcessed, new_prefix))
 	return;
     InitialiseAfterLoad(m_File, new_prefix);
 }
