@@ -4,7 +4,7 @@
 //  Main frame handling for Aven.
 //
 //  Copyright (C) 2000-2003,2005 Mark R. Shinwell
-//  Copyright (C) 2001-2003,2004,2005,2006,2010,2011,2012,2013,2014,2015 Olly Betts
+//  Copyright (C) 2001-2003,2004,2005,2006,2010,2011,2012,2013,2014,2015,2016 Olly Betts
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -65,6 +65,7 @@ enum {
     menu_FILE_PAGE_SETUP,
     menu_FILE_SCREENSHOT,
     menu_FILE_EXPORT,
+    menu_FILE_EXTEND,
     menu_PRES_NEW,
     menu_PRES_OPEN,
     menu_PRES_SAVE,
@@ -129,6 +130,8 @@ enum {
     menu_CTL_PERCENT,
     menu_CTL_REVERSE,
     menu_CTL_CANCEL_DIST_LINE,
+    menu_SURVEY_RESTRICT,
+    menu_SURVEY_SHOW_ALL,
     textctrl_FIND,
     button_HIDE,
     listctrl_PRES
@@ -207,6 +210,9 @@ class MainFrm : public wxFrame {
     wxNotebook* m_Notebook;
     AvenPresList* m_PresList;
     wxString m_File;
+    // Processed version of data - same as m_File if m_File is processed data.
+    wxString m_FileProcessed;
+    wxString m_Survey;
 public: // FIXME for m_cs_proj
     wxString m_Title, m_cs_proj, m_DateStamp;
 private:
@@ -252,7 +258,7 @@ public:
     bool LoadData(const wxString& file, const wxString& prefix);
     void AddToFileHistory(const wxString & file);
 
-    void InitialiseAfterLoad(const wxString & file);
+    void InitialiseAfterLoad(const wxString & file, const wxString & prefix);
     void OnShowLog(wxCommandEvent& event);
 
     void OnMRUFile(wxCommandEvent& event);
@@ -305,6 +311,7 @@ public:
     void OnPresStop(wxCommandEvent& event);
     void OnPresExportMovie(wxCommandEvent& event);
     void OnExport(wxCommandEvent& event);
+    void OnExtend(wxCommandEvent& event);
     void OnQuit(wxCommandEvent& event);
 
     void OnAbout(wxCommandEvent& event);
@@ -319,9 +326,13 @@ public:
 	}
     }
 
-    void OnShowLogUpdate(wxUpdateUIEvent &ui) { ui.Enable(m_Log != NULL); }
+    void OnShowLogUpdate(wxUpdateUIEvent &ui) {
+	ui.Enable(m_Log != NULL || (m_Splitter->GetWindow1() != m_Gfx && m_Splitter->GetWindow2() != m_Gfx));
+	ui.Check(m_Log == NULL);
+    }
     void OnPrintUpdate(wxUpdateUIEvent &ui) { ui.Enable(!m_File.empty()); }
     void OnExportUpdate(wxUpdateUIEvent &ui) { ui.Enable(!m_File.empty()); }
+    void OnExtendUpdate(wxUpdateUIEvent &ui) { ui.Enable(!m_IsExtendedElevation); }
 
     // temporary bodges until event handling problem is sorted out:
     void OnDefaultsUpdate(wxUpdateUIEvent& event) { if (m_Control) m_Control->OnDefaultsUpdate(event); }
@@ -535,6 +546,11 @@ public:
     PresentationMark GetPresMark(int which);
     bool Animating() const { return m_Gfx && m_Gfx->Animating(); }
 
+    // Restrict to subsurvey `survey` (or show all if `survey` empty).
+    void RestrictTo(const wxString & survey);
+
+    wxString GetSurvey() const { return m_Survey; }
+
 private:
     DECLARE_EVENT_TABLE()
 };
