@@ -295,7 +295,7 @@ static const unsigned format_info[] = {
     LABELS|LEGS|SURF|SPLAYS|ENTS|FIXES|EXPORTS|PROJ|EXPORT_3D,
     LABELS|LEGS|SURF|SPLAYS|STNS|CENTRED,
     LEGS|SPLAYS|CENTRED|EXPORT_3D,
-    LABELS|LEGS|SPLAYS|ENTS|FIXES|EXPORTS|PROJ|EXPORT_3D,
+    LABELS|LEGS|SPLAYS|PASG|XSECT|WALLS|ENTS|FIXES|EXPORTS|PROJ|EXPORT_3D,
     LABELS|LEGS|SURF|SPLAYS,
     LABELS|LEGS|SURF|SPLAYS|STNS|MARKER_SIZE|GRID|SCALE,
     LABELS|ENTS|FIXES|EXPORTS|EXPORT_3D,
@@ -339,6 +339,10 @@ static const int msg_filetype[] = {
     /*SVG files*/417
 };
 
+// We discriminate as "One Page" isn't valid for exporting.
+static wxString default_scale_print;
+static wxString default_scale_export;
+
 svxPrintDlg::svxPrintDlg(MainFrm* mainfrm_, const wxString & filename,
 			 const wxString & title, const wxString & cs_proj,
 			 const wxString & datestamp, time_t datestamp_numeric,
@@ -444,19 +448,28 @@ svxPrintDlg::svxPrintDlg(MainFrm* mainfrm_, const wxString & filename,
 
     wxStaticText* label;
     label = new wxStaticText(this, -1, wxString(wmsg(/*Scale*/154)) + wxT(" 1:"));
-    if (scales[0].empty()) {
-	if (printing) {
-	    /* TRANSLATORS: used in the scale drop down selector in the print
-	     * dialog the implicit meaning is "choose a suitable scale to fit
-	     * the plot on a single page", but we need something shorter */
-	    scales[0].assign(wmsg(/*One page*/258));
-	} else {
-	    scales[0].assign(wxT("1000"));
-	}
+    if (printing && scales[0].empty()) {
+	/* TRANSLATORS: used in the scale drop down selector in the print
+	 * dialog the implicit meaning is "choose a suitable scale to fit
+	 * the plot on a single page", but we need something shorter */
+	scales[0].assign(wmsg(/*One page*/258));
     }
-    m_scale = new wxComboBox(this, svx_SCALE, scales[0], wxDefaultPosition,
-			     wxDefaultSize, sizeof(scales) / sizeof(scales[0]),
-			     scales);
+    wxString default_scale;
+    if (printing) {
+	default_scale = default_scale_print;
+	if (default_scale.empty()) default_scale = scales[0];
+    } else {
+	default_scale = default_scale_export;
+	if (default_scale.empty()) default_scale = wxT("1000");
+    }
+    const wxString* scale_list = scales;
+    size_t n_scales = sizeof(scales) / sizeof(scales[0]);
+    if (!printing) {
+	++scale_list;
+	--n_scales;
+    }
+    m_scale = new wxComboBox(this, svx_SCALE, default_scale, wxDefaultPosition,
+			     wxDefaultSize, n_scales, scale_list);
     m_scalebox = new wxBoxSizer(wxHORIZONTAL);
     m_scalebox->Add(label, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);
     m_scalebox->Add(m_scale, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);
@@ -715,14 +728,15 @@ svxPrintDlg::OnExport(wxCommandEvent&) {
 	double marker_size = 0.8;
 
 	try {
-	    if (!Export(dlg.GetPath(), m_layout.title,
+	    const wxString& export_fnm = dlg.GetPath();
+	    if (!Export(export_fnm, m_layout.title,
 			m_layout.datestamp, m_layout.datestamp_numeric, mainfrm,
 			m_layout.rot, m_layout.tilt,
 			m_layout.get_effective_show_mask(),
 			export_format(format_idx), input_projection.utf8_str(),
 			grid, text_height, marker_size, m_layout.Scale)) {
 		wxString m = wxString::Format(wmsg(/*Couldn’t write file “%s”*/402).c_str(),
-					      m_File.c_str());
+					      export_fnm.c_str());
 		wxGetApp().ReportError(m);
 	    }
 	} catch (const wxString & m) {
@@ -800,6 +814,13 @@ svxPrintDlg::OnChangeSpin(wxSpinDoubleEvent& e) {
 
 void
 svxPrintDlg::OnChange(wxCommandEvent& e) {
+    if (e.GetId() == svx_SCALE && m_scale) {
+	default_scale_print = m_scale->GetValue();
+	if (default_scale_print != scales[0]) {
+	    // Don't store "One Page" for use when exporting.
+	    default_scale_export = default_scale_print;
+	}
+    }
     SomethingChanged(e.GetId());
 }
 
@@ -856,8 +877,8 @@ svxPrintDlg::SomethingChanged(int control_id) {
 	RecalcBounds();
 
 	if (m_scale) {
-	    (m_scale->GetValue()).ToDouble(&(m_layout.Scale));
-	    if (m_layout.Scale == 0.0) {
+	    if (!(m_scale->GetValue()).ToDouble(&(m_layout.Scale)) ||
+		m_layout.Scale == 0.0) {
 		m_layout.pick_scale(1, 1);
 	    }
 	}
@@ -870,28 +891,25 @@ svxPrintDlg::SomethingChanged(int control_id) {
 }
 
 void
-svxPrintDlg::LayoutToUI(){
+svxPrintDlg::LayoutToUI()
+{
 //    m_blanks->SetValue(m_layout.SkipBlank);
     if (m_layout.view != layout::EXTELEV) {
 	m_tilt->SetValue(m_layout.tilt);
 	m_bearing->SetValue(m_layout.rot);
     }
 
-    // Do this last as it causes an OnChange message which calls UIToLayout
-    if (m_scale) {
-	if (m_layout.Scale != 0) {
-	    wxString temp;
-	    temp << m_layout.Scale;
-	    m_scale->SetValue(temp);
-	} else {
-	    if (scales[0].empty()) scales[0].assign(wmsg(/*One page*/258));
-	    m_scale->SetValue(scales[0]);
-	}
+    if (m_scale && m_layout.Scale != 0) {
+	// Do this last as it causes an OnChange message which calls UIToLayout
+	wxString temp;
+	temp << m_layout.Scale;
+	m_scale->SetValue(temp);
     }
 }
 
 void
-svxPrintDlg::UIToLayout(){
+svxPrintDlg::UIToLayout()
+{
 //    m_layout.SkipBlank = m_blanks->IsChecked();
 
     if (m_layout.view != layout::EXTELEV && m_tilt) {
