@@ -1,6 +1,6 @@
 /* readval.c
  * Routines to read a prefix or number from the current input file
- * Copyright (C) 1991-2003,2005,2006,2010,2011,2012,2013,2014,2015 Olly Betts
+ * Copyright (C) 1991-2003,2005,2006,2010,2011,2012,2013,2014,2015,2016 Olly Betts
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -77,24 +77,26 @@ read_prefix(unsigned pfx_flags)
    bool fNew;
    bool fImplicitPrefix = fTrue;
    int depth = -1;
+   filepos fp_firstsep;
 
    skipblanks();
 #ifndef NO_DEPRECATED
    if (isRoot(ch)) {
       if (!(pfx_flags & PFX_ALLOW_ROOT)) {
-	 compile_error(-/*ROOT is deprecated*/25);
+	 compile_diagnostic(DIAG_ERR|DIAG_COL, /*ROOT is deprecated*/25);
 	 LONGJMP(file.jbSkipLine);
       }
       if (root_depr_count < 5) {
-	 compile_warning(-/*ROOT is deprecated*/25);
+	 compile_diagnostic(DIAG_WARN|DIAG_COL, /*ROOT is deprecated*/25);
 	 if (++root_depr_count == 5)
-	    compile_warning(/*Further uses of this deprecated feature will not be reported*/95);
+	    compile_diagnostic(DIAG_WARN, /*Further uses of this deprecated feature will not be reported*/95);
       }
       nextch();
       ptr = root;
       if (!isNames(ch)) {
 	 if (!isSep(ch)) return ptr;
 	 /* Allow optional SEPARATOR after ROOT */
+	 get_pos(&fp_firstsep);
 	 nextch();
       }
       fImplicitPrefix = fFalse;
@@ -116,7 +118,8 @@ read_prefix(unsigned pfx_flags)
 	     * the leg to it is a splay.
 	     */
 	    if (TSTBIT(pcs->flags, FLAGS_ANON_ONE_END)) {
-	       compile_error(-/*Can't have a leg between two anonymous stations*/3);
+	       set_pos(&here);
+	       compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Can't have a leg between two anonymous stations*/3);
 	       LONGJMP(file.jbSkipLine);
 	    }
 	    pcs->flags |= BIT(FLAGS_ANON_ONE_END) | BIT(FLAGS_IMPLICIT_SPLAY);
@@ -131,7 +134,8 @@ read_prefix(unsigned pfx_flags)
 	       prefix * pfx;
 anon_wall_station:
 	       if (TSTBIT(pcs->flags, FLAGS_ANON_ONE_END)) {
-		  compile_error(-/*Can't have a leg between two anonymous stations*/3);
+		  set_pos(&here);
+		  compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Can't have a leg between two anonymous stations*/3);
 		  LONGJMP(file.jbSkipLine);
 	       }
 	       pcs->flags |= BIT(FLAGS_ANON_ONE_END) | BIT(FLAGS_IMPLICIT_SPLAY);
@@ -148,7 +152,8 @@ anon_wall_station:
 		   * which isn't refindable).
 		   */
 		  if (TSTBIT(pcs->flags, FLAGS_ANON_ONE_END)) {
-		     compile_error(-/*Can't have a leg between two anonymous stations*/3);
+		     set_pos(&here);
+		     compile_diagnostic(DIAG_ERR|DIAG_TOKEN, /*Can't have a leg between two anonymous stations*/3);
 		     LONGJMP(file.jbSkipLine);
 		  }
 		  pcs->flags |= BIT(FLAGS_ANON_ONE_END);
@@ -186,19 +191,22 @@ anon_wall_station:
 	 }
 	 nextch();
       }
-      if (isSep(ch)) fImplicitPrefix = fFalse;
+      if (isSep(ch)) {
+	 fImplicitPrefix = fFalse;
+	 get_pos(&fp_firstsep);
+      }
       if (i == 0) {
 	 osfree(name);
 	 if (!f_optional) {
 	    if (isEol(ch)) {
 	       if (fSurvey) {
-		  compile_error(-/*Expecting survey name*/89);
+		  compile_diagnostic(DIAG_ERR|DIAG_COL, /*Expecting survey name*/89);
 	       } else {
-		  compile_error(-/*Expecting station name*/28);
+		  compile_diagnostic(DIAG_ERR|DIAG_COL, /*Expecting station name*/28);
 	       }
 	    } else {
 	       /* TRANSLATORS: Here "station" is a survey station, not a train station. */
-	       compile_error(-/*Character “%c” not allowed in station name (use *SET NAMES to set allowed characters)*/7, ch);
+	       compile_diagnostic(DIAG_ERR|DIAG_COL, /*Character “%c” not allowed in station name (use *SET NAMES to set allowed characters)*/7, ch);
 	    }
 	    LONGJMP(file.jbSkipLine);
 	 }
@@ -273,6 +281,7 @@ anon_wall_station:
       }
       depth++;
       f_optional = fFalse; /* disallow after first level */
+      if (isSep(ch)) get_pos(&fp_firstsep);
    } while (isSep(ch));
    if (name) osfree(name);
 
@@ -294,8 +303,8 @@ anon_wall_station:
 	  * Here "survey" is a "cave map" rather than list of questions - it should be
 	  * translated to the terminology that cavers using the language would use.
 	  */
-	 compile_error(/*“%s” can’t be both a station and a survey*/27,
-		       sprint_prefix(ptr));
+	 compile_diagnostic(DIAG_ERR, /*“%s” can’t be both a station and a survey*/27,
+			    sprint_prefix(ptr));
       }
       if (!fSurvey && TSTBIT(pcs->infer, INFER_EXPORTS)) ptr->min_export = USHRT_MAX;
    }
@@ -319,11 +328,11 @@ anon_wall_station:
       s = osstrdup(sprint_prefix(survey));
       p = sprint_prefix(ptr);
       if (survey->filename) {
-	 compile_error_pfx(survey,
-			   /*Station “%s” not exported from survey “%s”*/26,
-			   p, s);
+	 compile_diagnostic_pfx(DIAG_ERR, survey,
+				/*Station “%s” not exported from survey “%s”*/26,
+				p, s);
       } else {
-	 compile_error(/*Station “%s” not exported from survey “%s”*/26, p, s);
+	 compile_diagnostic(DIAG_ERR, /*Station “%s” not exported from survey “%s”*/26, p, s);
       }
       osfree(s);
 #if 0
@@ -333,7 +342,11 @@ anon_wall_station:
 #endif
    }
    if (!fImplicitPrefix && (pfx_flags & PFX_WARN_SEPARATOR)) {
-      compile_warning(/*Separator in survey name*/392);
+      filepos fp_tmp;
+      get_pos(&fp_tmp);
+      set_pos(&fp_firstsep);
+      compile_diagnostic(DIAG_WARN|DIAG_COL, /*Separator in survey name*/392);
+      set_pos(&fp_tmp);
    }
    return ptr;
 }
@@ -379,9 +392,9 @@ read_number(bool f_optional)
    }
 
    if (isOmit(ch_old)) {
-      compile_error(-/*Field may not be omitted*/8);
+      compile_diagnostic(DIAG_ERR|DIAG_COL, /*Field may not be omitted*/8);
    } else {
-      compile_error_token(-/*Expecting numeric field, found “%s”*/9);
+      compile_diagnostic_token_show(DIAG_ERR, /*Expecting numeric field, found “%s”*/9);
    }
    LONGJMP(file.jbSkipLine);
    return 0.0; /* for brain-fried compilers */
@@ -429,7 +442,7 @@ read_numeric_multi_or_omit(int *p_n_readings)
    real v = read_numeric_multi(fTrue, p_n_readings);
    if (v == HUGE_REAL) {
       if (!isOmit(ch)) {
-	 compile_error_token(-/*Expecting numeric field, found “%s”*/9);
+	 compile_diagnostic_token_show(DIAG_ERR, /*Expecting numeric field, found “%s”*/9);
 	 LONGJMP(file.jbSkipLine);
 	 return 0.0; /* for brain-fried compilers */
       }
@@ -445,7 +458,7 @@ read_uint_internal(int errmsg, const filepos *fp)
    unsigned int n = 0;
    if (!isdigit(ch)) {
       if (fp) set_pos(fp);
-      compile_error_token(-errmsg);
+      compile_diagnostic_token_show(DIAG_ERR, errmsg);
       LONGJMP(file.jbSkipLine);
    }
    while (isdigit(ch)) {
@@ -473,7 +486,7 @@ read_string(char **pstr, int *plen)
       nextch();
       while (1) {
 	 if (isEol(ch)) {
-	    compile_error(-/*Missing \"*/69);
+	    compile_diagnostic(DIAG_ERR|DIAG_COL, /*Missing \"*/69);
 	    LONGJMP(file.jbSkipLine);
 	 }
 
@@ -488,7 +501,7 @@ read_string(char **pstr, int *plen)
       while (1) {
 	 if (isEol(ch) || isComm(ch)) {
 	    if (!*pstr || !(*pstr)[0]) {
-	       compile_error(-/*Expecting string field*/121);
+	       compile_diagnostic(DIAG_ERR|DIAG_COL, /*Expecting string field*/121);
 	       LONGJMP(file.jbSkipLine);
 	    }
 	    return;
@@ -506,33 +519,39 @@ extern void
 read_date(int *py, int *pm, int *pd)
 {
    int y = 0, m = 0, d = 0;
-   filepos fp;
+   filepos fp_date;
 
    skipblanks();
 
-   get_pos(&fp);
-   y = read_uint_internal(/*Expecting date, found “%s”*/198, &fp);
+   get_pos(&fp_date);
+   y = read_uint_internal(/*Expecting date, found “%s”*/198, &fp_date);
    /* Two digit year is 19xx. */
    if (y < 100) y += 1900;
    if (y < 1900 || y > 2078) {
-      compile_warning(/*Invalid year (< 1900 or > 2078)*/58);
+      set_pos(&fp_date);
+      compile_diagnostic(DIAG_WARN|DIAG_UINT, /*Invalid year (< 1900 or > 2078)*/58);
       LONGJMP(file.jbSkipLine);
       return; /* for brain-fried compilers */
    }
    if (ch == '.') {
+      filepos fp;
       nextch();
-      m = read_uint_internal(/*Expecting date, found “%s”*/198, &fp);
+      get_pos(&fp);
+      m = read_uint_internal(/*Expecting date, found “%s”*/198, &fp_date);
       if (m < 1 || m > 12) {
-	 compile_warning(/*Invalid month*/86);
+	 set_pos(&fp);
+	 compile_diagnostic(DIAG_WARN|DIAG_UINT, /*Invalid month*/86);
 	 LONGJMP(file.jbSkipLine);
 	 return; /* for brain-fried compilers */
       }
       if (ch == '.') {
 	 nextch();
-	 d = read_uint_internal(/*Expecting date, found “%s”*/198, &fp);
+	 get_pos(&fp);
+	 d = read_uint_internal(/*Expecting date, found “%s”*/198, &fp_date);
 	 if (d < 1 || d > last_day(y, m)) {
+	    set_pos(&fp);
 	    /* TRANSLATORS: e.g. 31st of April, or 32nd of any month */
-	    compile_warning(/*Invalid day of the month*/87);
+	    compile_diagnostic(DIAG_WARN|DIAG_UINT, /*Invalid day of the month*/87);
 	    LONGJMP(file.jbSkipLine);
 	    return; /* for brain-fried compilers */
 	 }
